#!/usr/bin/env python3

"""Test that check config script handles various broken configs."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Innovation Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import shutil
import tempfile
import unittest
import re

from faucet.check_faucet_config import check_config


class CheckConfigTestCase(unittest.TestCase):  # pytype: disable=module-attr
    """Test that check config script handles various broken configs."""

    tmpdir = None

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def run_check_config(self, config, expected_ok):
        """Try to parse config and return True if result fails or succeeds as expected."""
        conf_files = []
        if config is not None:
            conf_file_name = os.path.join(self.tmpdir, 'faucet.yaml')
            with open(conf_file_name, 'w', encoding='utf-8') as conf_file:
                conf_file.write(config)
            conf_files = [conf_file_name]
        with open(os.devnull, 'w', encoding='utf-8') as check_output_file:
            result_ok = check_config(
                conf_files, logging.FATAL, check_output_file)
        return expected_ok == result_ok

    def _deprecated_acl_check(self, config, success):
        # TODO: Check acls_in work now acl_in is deprecated, remove in future
        if config and 'acl_in' in config and 'acls_in' not in config:
            acls_cfg = re.sub('(acl_in: )(.*)', 'acls_in: [\\2]', config)
            self.assertTrue(self.run_check_config(acls_cfg, success))

    def check_config_success(self, config):
        """Try to parse config and expect success."""
        self.assertTrue(self.run_check_config(config, True))
        self._deprecated_acl_check(config, True)

    def check_config_failure(self, config):
        """Try to parse config and expect failure."""
        self.assertTrue(self.run_check_config(config, False))
        self._deprecated_acl_check(config, False)

    def test_no_dps(self):
        """Test bad config with no DPs."""
        no_dps_conf = """
vlans:
    100:
        description: "100"
"""
        self.check_config_failure(no_dps_conf)

    def test_minimal(self):
        """Test minimal correct config."""
        minimal_conf = """
vlans:
    100:
        description: "100"
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
"""
        self.check_config_success(minimal_conf)

    def test_include_optional(self):
        """Test minimal include optional correct config."""
        minimal_conf = """
include-optional: ['/nonexistant']
vlans:
    100:
        description: "100"
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
"""
        self.check_config_success(minimal_conf)

    def test_include_required(self):
        """Test minimal include optional correct config."""
        minimal_conf = """
include: ['/nonexistant']
vlans:
    100:
        description: "100"
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
"""
        self.check_config_failure(minimal_conf)

    def test_invalid_vid(self):
        """Test invalid VID."""
        invalid_vid_conf = """
vlans:
    100:
        description: "100"
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 4097
"""
        self.check_config_failure(invalid_vid_conf)

    def test_vlan_name(self):
        """Test vlan referred by its name."""
        vlan_name_conf = """
vlans:
    finance:
        description: "FINANCE VLAN"
        vid: 100
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: finance
"""
        self.check_config_success(vlan_name_conf)

    def test_no_interfaces(self):
        """Test DP has no interfaces."""
        no_interfaces_conf = """
vlans:
    100:
        description: "100"
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
"""
        self.check_config_failure(no_interfaces_conf)

    def test_tabs(self):
        """Test that config with tabs is rejected."""
        tab_conf = """
vlans:
    100:
        	description: "100"  # noqa: E101,W191
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
"""
        self.check_config_failure(tab_conf)

    def test_no_vlan(self):
        """Test port without a VLAN rejected."""
        no_vlan_config = """
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'NOTSUPPORTED'
        interfaces:
            1:
                description: 'vlanless'
"""
        self.check_config_failure(no_vlan_config)

    def test_unknown_dp_config_item(self):
        """Test that an unknown DP field is rejected."""
        unknown_dp_config_item = """
vlans:
    100:
        description: "100"
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        broken: something
        interfaces:
            1:
                native_vlan: 100
"""
        self.check_config_failure(unknown_dp_config_item)

    def test_toplevel_unknown_config(self):
        """Test that an unknown toplevel config section is rejected."""
        toplevel_unknown_config = """
vlans:
    100:
        description: "100"
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
unknown_thing: 1
"""
        self.check_config_failure(toplevel_unknown_config)

    def test_toplevel_unknown_hardware(self):
        """Test that unknown hardware is rejected."""
        unknown_hardware_config = """
vlans:
    100:
        description: "100"
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'NOTSUPPORTED'
        interfaces:
            1:
                native_vlan: 100
"""
        self.check_config_failure(unknown_hardware_config)

    def test_bad_acl_action(self):
        """Test that an ACL with a bad match field is rejected."""
        acl_config = """
vlans:
    100:
        description: "100"
acls:
    101:
        - rule:
            nogood: '0e:00:00:00:02:02'
            actions:
                output:
                    port: 1
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
                acl_in: 101
"""
        self.check_config_failure(acl_config)

    def test_bad_acl_vlan(self):
        """Test that an ACL with a bad match field is rejected."""
        acl_config = """
vlans:
    100:
        description: "100"
        acl_in: 101
acls:
    101:
        - rule:
            nogood: "0e:00:00:00:02:02"
            actions:
                output:
                    port: 1
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
"""
        self.check_config_failure(acl_config)

    def test_good_acl(self):
        """Test that an ACL with good match field is accepted."""
        acl_config = """
vlans:
    100:
        description: "100"
acls:
    101:
        - rule:
            dl_dst: "0e:00:00:00:02:02"
            actions:
                output:
                    port: 1
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
                acl_in: 101
"""
        self.check_config_success(acl_config)

    def test_ports_good_acl(self):
        """Test that an ACL with good match field is accepted."""
        acl_config = """
vlans:
    100:
        description: "100"
acls:
    101:
        - rule:
            dl_dst: "0e:00:00:00:02:02"
            actions:
                output:
                    ports: [1, 2]
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
                acl_in: 101
            2:
                native_vlan: 100
                acl_in: 101
"""
        self.check_config_success(acl_config)

    def test_multiple_dps(self):
        """Test multiple DPs."""
        acl_config = """
vlans:
    100:
        description: "100"
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
    switch2:
        dp_id: 0xcafef00e
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
"""
        self.check_config_success(acl_config)

    def test_router_resolved_vlans(self):
        """Test that VLANs get resolved by routers."""
        vlan_config = """
vlans:
    100:
        description: "100"
    200:
        description: "200"
routers:
    router1:
        vlans: [100, 200]
dps:
    switch1:
        dp_id: 0xcafef00d
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 100
            2:
                native_vlan: 200
    switch2:
        dp_id: 0xdeadbeef
        hardware: 'Open vSwitch'
        interfaces:
            1:
                native_vlan: 200
"""
        self.check_config_success(vlan_config)

    def test_no_config_file(self):
        """Test no config file handled."""
        self.check_config_failure(None)


if __name__ == "__main__":
    unittest.main()  # pytype: disable=module-attr



"""RabbitMQ Adapter between FAUCET Events and RabbitMQ"""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import os
import select
import socket
import sys
import time

import pika


def get_sys_prefix():
    """This was copied from faucet.valve_util.
    Returns an additional prefix for log and configuration files when used in
    a virtual environment"""

    # Find the appropriate prefix for config and log file default locations
    # in case Faucet is run in a virtual environment. virtualenv marks the
    # original path in sys.real_prefix. If this value exists, and is
    # different from sys.prefix, then we are most likely running in a
    # virtualenv. Also check for Py3.3+ pyvenv.
    sysprefix = ''
    if (getattr(sys, 'real_prefix', sys.prefix) != sys.prefix
            or getattr(sys, 'base_prefix', sys.prefix) != sys.prefix):
        sysprefix = sys.prefix

    return sysprefix


# pylint: disable=too-many-instance-attributes
class RabbitAdapter:
    """A RabbitMQ adapter to get events from the FAUCET Unix socket and send
    them as messages to a RabbitMQ server
    """

    def __init__(self):
        super().__init__()

        # setup socket
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        # get environment variables and set defaults
        self.channel = None
        self.event_sock = os.getenv('FAUCET_EVENT_SOCK', '0')
        self.host = os.getenv('FA_RABBIT_HOST', '')
        self.port = os.getenv('FA_RABBIT_PORT')
        if not self.port:
            self.port = 5672
        else:
            try:
                self.port = int(self.port)
            except ValueError:
                self.port = 5672
        self.exchange = os.getenv('FA_RABBIT_EXCHANGE')
        if not self.exchange:
            self.exchange = 'topic_recs'
        self.exchange_type = os.getenv('FA_RABBIT_EXCHANGE_TYPE')
        if not self.exchange_type:
            self.exchange_type = 'topic'
        self.routing_key = os.getenv('FA_RABBIT_ROUTING_KEY', 'FAUCET.Event')
        if not self.routing_key:
            self.routing_key = 'FAUCET.Event'

    def rabbit_conn(self):
        """Make connection to rabbit to send events"""
        # check if a rabbit host was specified
        if not self.host:
            print('Not connecting to any RabbitMQ, host is None.')
            return False

        # create connection to rabbit
        params = pika.ConnectionParameters(host=self.host,
                                           port=self.port,
                                           heartbeat=600,
                                           blocked_connection_timeout=300)
        try:
            self.channel = pika.BlockingConnection(params).channel()
            self.channel.exchange_declare(exchange=self.exchange,
                                          exchange_type=self.exchange_type)
        except (pika.exceptions.AMQPError, socket.gaierror, OSError) as err:
            print(f"Unable to connect to RabbitMQ at {self.host}:{self.port} because: {err}")
            return False
        print(f"Connected to RabbitMQ at {self.host}:{self.port}")
        return True

    def socket_conn(self):
        """Make connection to sock to receive events"""
        # check if socket events are enabled
        if self.event_sock == '0':
            print('Not connecting to any socket, FAUCET_EVENT_SOCK is none.')
            return False
        if self.event_sock == '1':
            self.event_sock = get_sys_prefix() + '/var/run/faucet/faucet.sock'
        # otherwise it's a path

        # create connection to unix socket
        try:
            self.sock.connect(self.event_sock)
            self.sock.setblocking(False)
        except socket.error as err:
            print(f"Failed to connect to the socket because: {err}")
            return False
        print(f"Connected to the socket at {self.event_sock}")
        return True

    def poll_events(self):
        """Poll FAUCET socket for events."""
        socket_ok = False
        event_buffer = b''
        read_ready, _, _ = select.select([self.sock], [], [])
        if self.sock in read_ready:
            socket_ok = True
            while True:
                try:
                    event_buffer += self.sock.recv(1024)
                except socket.error as err:
                    socket_ok = err.errno == errno.EWOULDBLOCK
                    break

        events = event_buffer.strip().split(b'\n')
        return (socket_ok, events)

    def main(self):
        """Make connections to sock and rabbit and receive messages from sock
        to sent to rabbit
        """
        # ensure connections to the socket and rabbit before getting messages
        if self.rabbit_conn() and self.socket_conn():
            socket_ok = True
            events = []
            while socket_ok:
                if not events:
                    socket_ok, events = self.poll_events()
                try:
                    for event in events:
                        self.channel.basic_publish(
                            exchange=self.exchange,
                            routing_key=self.routing_key,
                            body=event,
                            properties=pika.BasicProperties(delivery_mode=2,))
                    events = []
                except pika.exceptions.AMQPError as err:
                    print(f"Unable to send events {events} to RabbitMQ: {err}, retrying")
                    time.sleep(1)
                    self.rabbit_conn()
                    sys.stdout.flush()
            self.sock.close()


if __name__ == "__main__":  # pragma: no cover
    RABBIT_ADAPTER = RabbitAdapter()
    RABBIT_ADAPTER.main()



#!/usr/bin/env python3

"""Mininet test runner

 * must be run as root
 * you can run a specific test case only, by adding the class name of the test
   case to the command. Eg ./mininet_test.py FaucetUntaggedIPv4RouteTest

It is strongly recommended to run these tests via Docker, to ensure you have
all dependencies correctly installed. See ../docs/.
"""

# pylint: disable=missing-class-docstring,missing-function-docstring

import argparse
import collections
import copy
import cProfile
import json
import glob
import inspect
import os
import sys
import pdb
import pstats
import random
import re
import shutil
import subprocess
import tempfile
import threading
import time
import traceback
import unittest

from packaging import version

from concurrencytest import ConcurrentTestSuite, fork_for_tests
from mininet.log import setLogLevel
from mininet.clean import Cleanup
import psutil

from clib import mininet_test_util
from clib.valve_test_lib import yaml_load, yaml_dump

DEFAULT_HARDWARE = 'Open vSwitch'

# Only these hardware types will be tested with meters.
SUPPORTS_METERS = (
    DEFAULT_HARDWARE,  # TODO: tested with OVS userspace only.
    'Aruba',
    'NoviFlow',
    'ZodiacGX',
)

SUPPORTS_METADATA = (
    DEFAULT_HARDWARE,
)


EXTERNAL_DEPENDENCIES = (
    ('osken-manager', ['--version'],
     'osken-manager', r'osken-manager\s+(\d+\.\d+\.\d+)\n', "2.1"),
    ('ovs-vsctl', ['--version'], 'Open vSwitch',
     r'ovs-vsctl\s+\(Open vSwitch\)\s+(\d+\.\d+)\.\d+\n', "2.3"),
    ('tcpdump', ['-h'], 'tcpdump',
     r'tcpdump\s+version\s+(\d+\.\d+)\.\d+\n', "4.5"),
    ('nc', ['-h'], 'OpenBSD netcat', '', 0),
    ('vconfig', [], 'the VLAN you are talking about', '', 0),
    ('fuser', ['-V'], r'fuser \(PSmisc\)',
     r'fuser \(PSmisc\) (\d+\.\d+|UNKNOWN)\n', "22.0"),
    ('lsof', ['-v'], r'lsof version',
     r'revision: (\d+\.\d+(\.\d+)?)\n', "4.86"),
    ('mn', ['--version'], r'\d+\.\d+.\d+',
     r'(\d+\.\d+).\d+', "2.2"),
    ('exabgp', ['--version'], 'ExaBGP',
     r'ExaBGP : (\d+\.\d+).\d+', "4.0"),
    ('pip3', ['show', 'influxdb'], 'influxdb',
     r'Version:\s+(\d+\.\d+)\.\d+', "3.0"),
    ('curl', ['--version'], 'libcurl',
     r'curl (\d+\.\d+).\d+', "7.3"),
    ('ladvd', ['-h'], 'ladvd',
     r'ladvd version (\d+\.\d+)\.\d+', "0.9"),
    ('iperf', ['--version'], 'iperf',
     r'iperf version (\d+\.\d+)\.\d+', "2.0"),
    ('fping', ['-v'], 'fping',
     r'fping: Version (\d+\.\d+)', "3.10"),
    ('rdisc6', ['-V'], 'ndisc6',
     r'ndisc6.+tool (\d+\.\d+)', "1.0"),
    ('tshark', ['-v'], r'(?i)tshark',
     r'TShark\s*[a-zA-Z\(\)]*\s*([\d\.]+)', "2.1"),
    ('scapy', ['-h'], 'Usage: scapy', '', 0),
)

# see hw_switch_config.yaml for how to bridge in an external hardware switch.
HW_SWITCH_CONFIG_FILE = 'hw_switch_config.yaml'
CONFIG_FILE_DIRS = ['/etc/faucet', './', '/faucet-src']
REQUIRED_TEST_PORTS = 4

REQUIRED_HW_CONFIG = {
    'dp_ports': (dict,),
    'cpn_intf': (str,),
    'dpid': (int,),
    'hardware': (str,),
    'hw_switch': (bool,),
    'of_port': (int,),
    'gauge_of_port': (int,),
}

OPTIONAL_HW_CONFIG = {
    'cpn_ipv6': (bool,),
    'ctl_privkey': (str,),
    'ca_certs': (str,),
    'ctl_cert': (str,),
}

ALL_HW_CONFIG = {**REQUIRED_HW_CONFIG, **OPTIONAL_HW_CONFIG}


def import_hw_config():
    """Import configuration for physical switch testing."""
    for config_file_dir in CONFIG_FILE_DIRS:
        config_file_name = os.path.join(config_file_dir, HW_SWITCH_CONFIG_FILE)
        if os.path.isfile(config_file_name):
            break
    if os.path.isfile(config_file_name):
        print(f'Using config from {config_file_name}')
    else:
        print(f'Cannot find {HW_SWITCH_CONFIG_FILE} in {CONFIG_FILE_DIRS}')
        sys.exit(-1)
    try:
        with open(config_file_name, 'r', encoding='utf-8') as config_file:
            config = yaml_load(config_file)
    except IOError:
        print(f'Could not load YAML config data from {config_file_name}')
        sys.exit(-1)
    if config.get('hw_switch', False):
        unknown_keys = set(config.keys()) - set(ALL_HW_CONFIG.keys())
        if unknown_keys:
            print(f'unknown config {unknown_keys} in {config_file_name}')
            sys.exit(-1)
        missing_required_keys = set(REQUIRED_HW_CONFIG.keys()) - set(config.keys())
        if missing_required_keys:
            print(f'missing required config: {missing_required_keys}')
            sys.exit(-1)
        for config_key, config_value in config.items():
            valid_types = ALL_HW_CONFIG[config_key]
            valid_values = [
                config_value for valid_type in valid_types
                if isinstance(config_value, valid_type)]
            if not valid_values:
                print(f'{config_key} ({config_value}) must be of type {valid_types} in {config_file_name}')
                sys.exit(-1)
        dp_ports = config['dp_ports']
        if len(dp_ports) < REQUIRED_TEST_PORTS:
            print(f'At least {REQUIRED_TEST_PORTS} dataplane ports are required, '
                  f'{len(dp_ports)} are provided in {config_file_name}.')
            sys.exit(-1)
        return config
    return None


def check_dependencies():
    """Verify dependant libraries/binaries are present with correct versions."""
    print('Checking library/binary dependencies')
    for (binary, binary_get_version, binary_present_re,
         binary_version_re, binary_minversion) in EXTERNAL_DEPENDENCIES:
        binary_args = [binary] + binary_get_version
        required_binary = f'required binary/library {" ".join(binary_args)}'
        try:
            with subprocess.Popen(
                    binary_args,
                    stdin=mininet_test_util.DEVNULL,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    close_fds=True) as proc:
                proc_out, proc_err = proc.communicate()
                binary_output = proc_out.decode()
                if proc_err is not None:
                    binary_output += proc_err.decode()
        except subprocess.CalledProcessError:
            # Might have run successfully, need to parse output
            pass
        except OSError:
            print(f'could not run {required_binary}')
            return False
        present_match = re.search(binary_present_re, binary_output)
        if not present_match:
            print(f'{required_binary} not present or did not return expected string '
                  f'{binary_present_re} ({binary_output})')
            return False
        if binary_version_re:
            version_match = re.search(binary_version_re, binary_output)
            if version_match is None:
                print(f'could not get version from {required_binary} ({binary_output})')
                return False
            try:
                binary_version = version_match.group(1)
            except ValueError:
                print(f'cannot parse version {version_match} for {required_binary}')
                return False
            if binary == 'fuser' and binary_version == 'UNKNOWN':
                # Workaround for psmisc 23.3
                return True
            if version.parse(binary_version) < version.parse(binary_minversion):
                print(f'{required_binary} version {binary_version} is less than required version {binary_minversion}')
                return False
    return True


def make_suite(tc_class, hw_config, root_tmpdir, ports_sock, max_test_load,
               port_order, start_port):
    """Compose test suite based on test class names."""
    testloader = unittest.TestLoader()
    testnames = testloader.getTestCaseNames(tc_class)
    suite = unittest.TestSuite()
    for name in testnames:
        suite.addTest(tc_class(
            name, hw_config, root_tmpdir, ports_sock, max_test_load,
            port_order, start_port))
    return suite


def pipeline_superset_report(decoded_pcap_logs):
    """Report on matches, instructions, and actions by table from tshark logs."""

    def parse_flow(flow_lines):
        table_id = None
        group_id = None
        last_oxm_match = ''
        matches_count = 0
        actions_count = 0
        instructions_count = 0
        oxm_match_re = re.compile(r'.*Field: (\S+).*')
        oxm_mask_match_re = re.compile(r'.*Has mask: True.*')
        type_match_re = re.compile(r'Type: (\S+).+')

        for flow_line, depth, section_stack in flow_lines:
            if depth == 1:
                if flow_line.startswith('Type: OFPT_'):
                    if not (flow_line.startswith('Type: OFPT_FLOW_MOD')
                            or flow_line.startswith('Type: OFPT_GROUP_MOD')):
                        return
                if flow_line.startswith('Table ID'):
                    if not flow_line.startswith('Table ID: OFPTT_ALL'):
                        table_id = int(flow_line.split()[-1])
                    else:
                        return
                if flow_line.startswith('Group ID'):
                    if not flow_line.startswith('Group ID: OFPG_ALL'):
                        group_id = int(flow_line.split()[-1])
                    else:
                        return
                continue
            if depth > 1:
                section_name = section_stack[-1]
                if table_id is not None:
                    if 'Match' in section_stack:
                        if section_name == 'OXM field':
                            oxm_match = oxm_match_re.match(flow_line)
                            if oxm_match:
                                table_matches[table_id].add(oxm_match.group(1))
                                last_oxm_match = oxm_match.group(1)
                                matches_count += 1
                                if matches_count > table_matches_max[table_id]:
                                    table_matches_max[table_id] = matches_count
                            else:
                                oxm_mask_match = oxm_mask_match_re.match(flow_line)
                                if oxm_mask_match:
                                    table_matches[table_id].add(last_oxm_match + '/Mask')
                    elif 'Instruction' in section_stack:
                        type_match = type_match_re.match(flow_line)
                        if type_match:
                            if section_name == 'Instruction':
                                table_instructions[table_id].add(type_match.group(1))
                                instructions_count += 1
                                if instructions_count > table_instructions_max[table_id]:
                                    table_instructions_max[table_id] = instructions_count
                            elif section_name == 'Action':
                                table_actions[table_id].add(type_match.group(1))
                                actions_count += 1
                                if actions_count > table_actions_max[table_id]:
                                    table_actions_max[table_id] = actions_count
                elif group_id is not None:
                    if 'Bucket' in section_stack:
                        type_match = type_match_re.match(flow_line)
                        if type_match:
                            if section_name == 'Action':
                                group_actions.add(type_match.group(1))

    group_actions = set()
    table_matches = collections.defaultdict(set)
    table_matches_max = collections.defaultdict(lambda: 0)
    table_instructions = collections.defaultdict(set)
    table_instructions_max = collections.defaultdict(lambda: 0)
    table_actions = collections.defaultdict(set)
    table_actions_max = collections.defaultdict(lambda: 0)

    for log in decoded_pcap_logs:
        with open(log, encoding='utf-8') as log_file:
            packets = re.compile(r'\n{2,}').split(log_file.read())
        for packet in packets:
            last_packet_line = None
            indent_count = 0
            last_indent_count = 0
            section_stack = []
            flow_lines = []

            for packet_line in packet.splitlines():
                orig_packet_line = len(packet_line)
                packet_line = packet_line.lstrip()
                indent_count = orig_packet_line - len(packet_line)
                if indent_count == 0:
                    parse_flow(flow_lines)
                    flow_lines = []
                    section_stack = []
                elif indent_count > last_indent_count:
                    section_stack.append(last_packet_line)
                elif indent_count < last_indent_count:
                    if section_stack:
                        section_stack.pop()
                depth = len(section_stack)
                last_indent_count = indent_count
                last_packet_line = packet_line
                flow_lines.append((packet_line, depth, copy.copy(section_stack)))
            parse_flow(flow_lines)

    for table in sorted(table_matches):
        print(f'table: {table}')
        print(f'  matches: {sorted(table_matches[table])} (max {table_matches_max[table]})')
        print(f'  table_instructions: {sorted(table_instructions[table])} (max {table_instructions_max[table]})')
        print(f'  table_actions: {sorted(table_actions[table])} (max {table_actions_max[table]})')
    if group_actions:
        print('group bucket actions:')
        print(f'  {sorted(group_actions)}')


def filter_test_hardware(test_obj, hw_config):
    test_hosts = test_obj.N_TAGGED + test_obj.N_UNTAGGED + test_obj.N_EXTENDED
    test_links = test_hosts * test_obj.LINKS_PER_HOST
    testing_hardware = hw_config is not None
    test_hardware = DEFAULT_HARDWARE
    if testing_hardware:
        test_hardware = hw_config['hardware']

    if test_obj.REQUIRES_METERS and test_hardware not in SUPPORTS_METERS:
        return False

    if test_obj.REQUIRES_METADATA and test_hardware not in SUPPORTS_METADATA:
        return False

    if testing_hardware:
        if test_obj.SOFTWARE_ONLY:
            return False
        if test_obj.NUM_DPS > 1:
            # TODO: test other stacking combinations.
            if test_obj.NUM_HOSTS > 2:
                return False

    if test_obj.NUM_DPS == 1 and test_links < REQUIRED_TEST_PORTS:
        return False

    return True


def max_loadavg():
    return int(psutil.cpu_count(logical=False) * 1.5)


def expand_tests(modules, requested_test_classes, regex_test_classes, excluded_test_classes,
                 hw_config, root_tmpdir, ports_sock, serial,
                 port_order, start_port):
    sanity_test_suites = []
    single_test_suites = []
    parallel_test_suites = []

    for module in modules:
        for full_name, test_obj in inspect.getmembers(sys.modules[module]):
            test_name = full_name.split('.')[-1]
            if not inspect.isclass(test_obj):
                continue
            if regex_test_classes or requested_test_classes:
                if not ((regex_test_classes and regex_test_classes.match(test_name)) or (
                        requested_test_classes and test_name in requested_test_classes)):
                    continue
            if excluded_test_classes and test_name in excluded_test_classes:
                continue
            if test_name.endswith('Test') and test_name.startswith('Faucet'):
                if not filter_test_hardware(test_obj, hw_config):
                    continue
                print(f'adding test {test_name}')
                test_suite = make_suite(
                    test_obj, hw_config, root_tmpdir, ports_sock, max_loadavg(),
                    port_order, start_port)
                if test_name.startswith('FaucetSanity'):
                    sanity_test_suites.append(test_suite)
                else:
                    if serial or test_name.startswith('FaucetSingle') or test_obj.NETNS:
                        single_test_suites.append(test_suite)
                    else:
                        parallel_test_suites.append(test_suite)

        sanity_tests = unittest.TestSuite()
        single_tests = unittest.TestSuite()
        parallel_tests = unittest.TestSuite()

        if len(parallel_test_suites) == 1:
            single_test_suites.extend(parallel_test_suites)
            parallel_test_suites = []
        if parallel_test_suites:
            seed = time.time()
            print(f'seeding parallel test shuffle with {seed}')
            random.seed(seed)
            random.shuffle(parallel_test_suites)
            for test_suite in parallel_test_suites:
                parallel_tests.addTest(test_suite)

        for test_suite in sanity_test_suites:
            sanity_tests.addTest(test_suite)
        for test_suite in single_test_suites:
            single_tests.addTest(test_suite)
    return (sanity_tests, single_tests, parallel_tests)


class FaucetResult(unittest.runner.TextTestResult):  # pytype: disable=module-attr

    root_tmpdir = None
    test_duration_secs = {}
    unexpected_success = False

    def _test_tmpdir(self, test):
        return os.path.join(
            self.root_tmpdir, mininet_test_util.flat_test_name(test.id()))

    def _set_test_duration_secs(self, test):
        duration_file_name = os.path.join(self._test_tmpdir(test), 'test_duration_secs')
        if test.id() not in self.test_duration_secs:
            self.test_duration_secs[test.id()] = 0
        try:
            with open(duration_file_name, encoding='utf-8') as duration_file:
                self.test_duration_secs[test.id()] = int(duration_file.read())
        except FileNotFoundError:
            pass

    def stopTest(self, test):
        self._set_test_duration_secs(test)
        super().stopTest(test)

    def addUnexpectedSuccess(self, test):
        self.unexpected_success = True
        super().addUnexpectedSuccess(test)


class FaucetCleanupResult(FaucetResult):

    successes = []

    def addSuccess(self, test):
        self._set_test_duration_secs(test)
        shutil.rmtree(self._test_tmpdir(test))
        self.successes.append((test, ''))
        super().addSuccess(test)


def debug_exception_handler(etype, value, trace):
    traceback.print_exception(etype, value, trace)
    print()
    pdb.pm()


def test_runner(root_tmpdir, resultclass, failfast=False):
    resultclass.root_tmpdir = root_tmpdir
    return unittest.TextTestRunner(verbosity=255, resultclass=resultclass, failfast=failfast)


def run_parallel_test_suites(root_tmpdir, resultclass, parallel_tests):
    results = []
    if parallel_tests.countTestCases():
        max_parallel_tests = min(parallel_tests.countTestCases(), max_loadavg())
        print(f'running maximum of {max_parallel_tests} parallel tests')
        parallel_runner = test_runner(root_tmpdir, resultclass)
        parallel_suite = ConcurrentTestSuite(
            parallel_tests, fork_for_tests(max_parallel_tests))
        results.append(parallel_runner.run(parallel_suite))
    return results


def run_single_test_suites(debug, root_tmpdir, resultclass, single_tests):
    results = []
    # TODO: Tests that are serialized generally depend on hardcoded ports.
    # Make them use dynamic ports.
    if single_tests.countTestCases():
        single_runner = test_runner(root_tmpdir, resultclass)
        if debug:
            oldexcepthook = sys.excepthook
            sys.excepthook = debug_exception_handler
            single_tests.debug()
            sys.excepthook = oldexcepthook
        else:
            results.append(single_runner.run(single_tests))
    return results


def run_sanity_test_suite(root_tmpdir, resultclass, sanity_tests):
    sanity_runner = test_runner(root_tmpdir, resultclass, failfast=True)
    sanity_result = sanity_runner.run(sanity_tests)
    return sanity_result


def report_tests(test_status, test_list, result):
    tests_json = {}
    for test_class, test_text in test_list:
        test_text = test_text.replace('\n', '\t')
        test_text = test_text.replace('"', '\'')
        test_duration_secs = result.test_duration_secs[test_class.id()]
        tests_json.update({
            test_class.id(): {
                'status': test_status,
                'output': test_text,
                'test_duration_secs': test_duration_secs
            }})
    return tests_json


def report_results(results, hw_config, report_json_filename):
    if results:
        tests_json = {}
        report_title = 'test results'
        print('\n')
        print(report_title)
        print('=' * len(report_title))
        for result in results:
            test_lists = [
                ('ERROR', result.errors),
                ('FAIL', result.failures),
                ('SKIPPED', result.skipped),
            ]
            if hasattr(result, 'successes'):
                test_lists.append(
                    ('OK', result.successes))
            for test_status, test_list in test_lists:
                tests_json.update(report_tests(test_status, test_list, result))
        print(yaml_dump(tests_json))
        if report_json_filename:
            report_json = {
                'hw_config': hw_config,
                'tests': tests_json,
            }
            with open(report_json_filename, 'w', encoding='utf-8') as report_json_file:
                report_json_file.write(json.dumps(report_json))


def run_test_suites(debug, report_json_filename, hw_config, root_tmpdir,
                    resultclass, single_tests, parallel_tests, sanity_result):
    print(f'running {parallel_tests.countTestCases()} tests in parallel and '
          f'{single_tests.countTestCases()} tests serial')
    results = []
    results.append(sanity_result)
    results.extend(run_parallel_test_suites(root_tmpdir, resultclass, parallel_tests))
    results.extend(run_single_test_suites(debug, root_tmpdir, resultclass, single_tests))
    report_results(results, hw_config, report_json_filename)
    successful_results = [result for result in results
                          if result.wasSuccessful() or result.unexpected_success]
    return len(results) == len(successful_results)


def start_port_server(root_tmpdir, start_free_ports, min_free_ports):
    ports_sock = os.path.join(root_tmpdir, '.ports-server')
    ports_server = threading.Thread(
        target=mininet_test_util.serve_ports,
        args=(ports_sock, start_free_ports, min_free_ports))
    ports_server.setDaemon(True)
    ports_server.start()
    for _ in range(min_free_ports // 2):
        if os.path.exists(ports_sock):
            break
        time.sleep(1)
    if not os.path.exists(ports_sock):
        print(f'ports server did not start ({ports_sock} not created)')
        sys.exit(-1)
    return ports_sock


def dump_failed_test_file(test_file, only_exts):
    dump_file = False
    if only_exts:
        for ext in only_exts:
            if test_file.endswith(ext):
                dump_file = True
                break
    else:
        dump_file = True

    if dump_file:
        try:
            with open(test_file, encoding='utf-8') as test_file_h:
                test_file_content = test_file_h.read()
            if test_file_content:
                print(test_file)
                print('=' * len(test_file))
                print('\n')
                print(test_file_content)
        except UnicodeDecodeError:
            pass
    return dump_file


def dump_failed_test(test_name, test_dir):
    print(test_name)
    print('=' * len(test_name))
    print('\n')
    test_files = set(glob.glob(os.path.join(test_dir, '*')))
    dumped_test_files = set()

    for only_exts in (['.yaml'], ['.log'], ['.cap.txt'], ['.txt']):
        for test_file in sorted(test_files):
            if test_file in dumped_test_files:
                continue
            if dump_failed_test_file(test_file, only_exts):
                dumped_test_files.add(test_file)


def clean_test_dirs(root_tmpdir, all_successful, sanity, keep_logs, dumpfail):
    if all_successful:
        if not keep_logs or not os.listdir(root_tmpdir):
            shutil.rmtree(root_tmpdir)
    else:
        print(f'\nlog/debug files for failed tests are in {root_tmpdir}\n')
        if not keep_logs:
            if sanity:
                test_dirs = glob.glob(os.path.join(root_tmpdir, '*'))
                for test_dir in test_dirs:
                    test_name = os.path.basename(test_dir)
                    if dumpfail:
                        dump_failed_test(test_name, test_dir)


def run_tests(modules, hw_config, requested_test_classes, regex_test_classes, dumpfail, debug,
              keep_logs, serial, repeat, excluded_test_classes, report_json_filename,
              port_order, start_port):
    """Actually run the test suites, potentially in parallel."""
    if repeat:
        print('Will repeat tests until failure')
    if hw_config is not None:
        print('Testing hardware, forcing test serialization')
        serial = True
    root_tmpdir = tempfile.mkdtemp(prefix='faucet-tests-', dir='/var/tmp')
    os.chmod(root_tmpdir, 0o755)
    print(f'Logging test results in {root_tmpdir}')
    start_free_ports = 10
    min_free_ports = 200
    if serial:
        start_free_ports = 5
        min_free_ports = 5
    ports_sock = start_port_server(root_tmpdir, start_free_ports, min_free_ports)
    print('test ports server started')
    resultclass = FaucetCleanupResult
    if keep_logs:
        resultclass = FaucetResult
    all_successful = False
    no_tests = True

    sanity_tests, single_tests, parallel_tests = expand_tests(
        modules, requested_test_classes, regex_test_classes, excluded_test_classes,
        hw_config, root_tmpdir, ports_sock, serial, port_order, start_port)

    test_count = (sanity_tests.countTestCases() + single_tests.countTestCases() + parallel_tests.countTestCases())

    if test_count:
        no_tests = False
        sanity_result = run_sanity_test_suite(root_tmpdir, resultclass, sanity_tests)
        if sanity_result.wasSuccessful():
            while True:
                all_successful = run_test_suites(
                    debug, report_json_filename,
                    hw_config, root_tmpdir, resultclass,
                    copy.deepcopy(single_tests),
                    copy.deepcopy(parallel_tests),
                    sanity_result)
                if not repeat:
                    break
                if not all_successful:
                    break
                print('repeating run')
        else:
            report_results([sanity_result], hw_config, report_json_filename)

    if no_tests:
        print('no tests selected')
        shutil.rmtree(root_tmpdir)
        sys.exit(0)
    else:
        decoded_pcap_logs = glob.glob(os.path.join(
            os.path.join(root_tmpdir, '*'), '*of.cap.txt'))
        pipeline_superset_report(decoded_pcap_logs)
        clean_test_dirs(
            root_tmpdir, all_successful,
            sanity_result.wasSuccessful(), keep_logs, dumpfail)

    if not all_successful:
        sys.exit(-1)


def parse_args():
    """Parse command line arguments."""

    parser = argparse.ArgumentParser(
        prog='mininet_tests')
    parser.add_argument(
        '--regex', help='run tests that match regular expression pattern')
    parser.add_argument(
        '-c', '--clean', action='store_true', help='run mininet cleanup')
    parser.add_argument(
        '-d', '--dumpfail', action='store_true', help='dump logs for failed tests')
    parser.add_argument(
        '--debug', action='store_true', help='enter debug breakpoint on assertion failure')
    parser.add_argument(
        '-i', '--integration', default=True, action='store_true', help='run integration tests')
    parser.add_argument(
        '-j', '--jsonreport', help='write a json file with test results')
    parser.add_argument(
        '-k', '--keep_logs', action='store_true', help='keep logs even for OK tests')
    loglevels = ('debug', 'error', 'warning', 'info', 'output')
    parser.add_argument(
        '-l', '--loglevel', choices=loglevels, default='warning',
        help='set mininet logging level')
    parser.add_argument(
        '-n', '--nocheck', action='store_true', help='skip dependency check')
    parser.add_argument(
        '-o', '--order', default='random',
        help='port order for tests: 0,1,2,3 | random (default: random)')
    parser.add_argument(
        '-p', '--profile', action='store_true',
        help='use Cprofile to report elapsed wall clock time per function')
    parser.add_argument(
        '--port', default='random',
        help='starting port number (integer) | random (default: random)')
    parser.add_argument(
        '-r', '--repeat', action='store_true', help='repeat tests until failure')
    parser.add_argument(
        '-s', '--serial', action='store_true', help='run tests serially')
    parser.add_argument(
        '--generative_unit', default=False, action='store_true', help='run generative unit tests')
    parser.add_argument(
        '--generative_tolerance', default=False, action='store_true',
        help='run generative integration fault-tolerance tests')
    parser.add_argument(
        '-x', help='list of test classes to exclude')

    excluded_test_classes = []
    report_json_filename = None

    try:
        args, requested_test_classes = parser.parse_known_args(sys.argv[1:])
        if args.order == 'random':
            port_order = list(range(4))
            random.shuffle(port_order)
        else:
            port_order = [int(s) for s in args.order.split(',')]
        if sorted(port_order) != sorted(range(len(port_order))):
            print('Port order should be a permutation of 0,1,2,3')
            raise ValueError
        if args.port == 'random':
            start_port = random.randint(1, 10)
        else:
            start_port = int(args.port)
        regex_test_classes = None
        if args.regex:
            regex_test_classes = re.compile(args.regex)
            print(f'Running tests on classes matching {args.regex}')
    except(KeyError, IndexError, ValueError):
        parser.print_usage()
        sys.exit(-1)

    if args.jsonreport:
        report_json_filename = args.jsonreport
    if args.x:
        excluded_test_classes = args.x.split(',')

    return (
        requested_test_classes, regex_test_classes, args.clean, args.dumpfail, args.debug,
        args.keep_logs, args.nocheck, args.serial, args.repeat,
        excluded_test_classes, report_json_filename, port_order, start_port,
        args.loglevel, args.profile)


def test_main(modules, serial_override=None):
    """Test main."""

    print(f'testing module {modules}')

    (requested_test_classes, regex_test_classes, clean, dumpfail, debug, keep_logs, nocheck,
     serial, repeat, excluded_test_classes, report_json_filename, port_order,
     start_port, loglevel, profile) = parse_args()

    if serial_override is not None:
        print('overriding serial to ', serial_override)
        serial = serial_override

    setLogLevel(loglevel)

    if clean:
        print('Cleaning up test interfaces, processes and openvswitch '
              'configuration from previous test runs')
        Cleanup.cleanup()
        sys.exit(0)

    if nocheck:
        print('Skipping dependency checks')
    else:
        if not check_dependencies():
            print('dependency check failed. check required library/binary '
                  'list in header of this script')
            sys.exit(-1)

    print('port order: -o', ','.join(str(i) for i in port_order))
    print(f'start port: --port {start_port}')

    hw_config = import_hw_config()

    if profile:
        # use wall clock time
        pr = cProfile.Profile(time.time)  # pylint: disable=invalid-name
        pr.enable()

    run_tests(
        modules, hw_config, requested_test_classes, regex_test_classes, dumpfail, debug,
        keep_logs, serial, repeat, excluded_test_classes, report_json_filename,
        port_order, start_port)

    if profile:
        pr.disable()
        ps = pstats.Stats(pr).sort_stats('cumulative')  # pylint: disable=invalid-name
        ps.print_stats()



"""Manage LLDP."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict

from faucet import valve_util
from faucet import valve_of
from faucet import valve_packet
from faucet.valve_manager_base import ValveManagerBase


class ValveLLDPManager(ValveManagerBase):
    """Manage LLDP."""

    def __init__(self, vlan_table, highest_priority, logger,
                 notify, inc_var, set_var, set_port_var, stack_manager):
        self.vlan_table = vlan_table
        self.highest_priority = highest_priority
        self.logger = logger
        self.notify = notify
        self._set_var = set_var
        self._inc_var = inc_var
        self._set_port_var = set_port_var
        self.stack_manager = stack_manager

    def _lldp_match(self, port):
        return self.vlan_table.match(
            in_port=port.number,
            eth_dst=valve_packet.LLDP_MAC_NEAREST_BRIDGE,
            eth_dst_mask=valve_packet.BRIDGE_GROUP_MASK,
            eth_type=valve_of.ether.ETH_TYPE_LLDP)

    def add_port(self, port):
        ofmsgs = []
        if port.receive_lldp:
            ofmsgs.append(self.vlan_table.flowcontroller(
                match=self._lldp_match(port),
                priority=self.highest_priority,
                max_len=128))
        return ofmsgs

    def del_port(self, port):
        ofmsgs = []
        if port.receive_lldp:
            ofmsgs.append(self.vlan_table.flowdel(
                match=self._lldp_match(port),
                priority=self.highest_priority))
        return ofmsgs

    def verify_lldp(self, port, now, valve, other_valves,
                    remote_dp_id, remote_dp_name,
                    remote_port_id, remote_port_state):
        """
        Verify correct LLDP cabling, then update port to next state

        Args:
            port (Port): Port that received the LLDP
            now (float): Current time
            other_valves (list): Other valves in the topology
            remote_dp_id (int): Received LLDP remote DP ID
            remote_dp_name (str): Received LLDP remote DP name
            remote_port_id (int): Recevied LLDP port ID
            remote_port_state (int): Received LLDP port state
        Returns:
            dict: Ofmsgs by valve
        """
        if not port.stack:
            return {}
        remote_dp = port.stack['dp']
        remote_port = port.stack['port']
        stack_correct = True
        self._inc_var('stack_probes_received')
        if (remote_dp_id != remote_dp.dp_id
                or remote_dp_name != remote_dp.name
                or remote_port_id != remote_port.number):
            self.logger.error(
                f'Stack {port} cabling incorrect, expected '
                f'{valve_util.dpid_log(remote_dp.dp_id)}:{remote_dp.name}:{remote_port.number}, '
                f'actual {valve_util.dpid_log(remote_dp_id)}:{remote_dp_name}:{remote_port_id}')
            stack_correct = False
            self._inc_var('stack_cabling_errors')
        port.dyn_stack_probe_info = {
            'last_seen_lldp_time': now,
            'stack_correct': stack_correct,
            'remote_dp_id': remote_dp_id,
            'remote_dp_name': remote_dp_name,
            'remote_port_id': remote_port_id,
            'remote_port_state': remote_port_state
        }
        return self.update_stack_link_state([port], now, valve, other_valves)

    def update_stack_link_state(self, ports, now, valve, other_valves):
        """
        Update the stack link states of the set of provided stack ports

        Args:
            ports (list): List of stack ports to update the state of
            now (float): Current time
            valve (Valve): Valve that owns this LLDPManager instance
            other_valves (list): List of other valves
        Returns:
            dict: ofmsgs by valve
        """
        stack_changes = 0
        ofmsgs_by_valve = defaultdict(list)
        stacked_valves = set()
        if self.stack_manager:
            stacked_valves = {valve}.union(self.stack_manager.stacked_valves(other_valves))
        for port in ports:
            before_state = port.stack_state()
            after_state, reason = port.stack_port_update(now)
            if before_state != after_state:
                self._set_port_var('port_stack_state', after_state, port)
                self._inc_var(
                    'port_stack_state_change_count', labels=valve.dp.port_labels(port.number))
                self.notify({'STACK_STATE': {
                    'port': port.number,
                    'state': after_state}})
                self.logger.info(f'Stack {port} state {port.stack_state_name(after_state)} '
                                 f'(previous state {port.stack_state_name(before_state)}): {reason}')
                stack_changes += 1
                port_up = False
                if port.is_stack_up():
                    port_up = True
                elif port.is_stack_init() and port.stack['port'].is_stack_up():
                    port_up = True
                for stack_valve in stacked_valves:
                    stack_valve.stack_manager.update_stack_topo(port_up, valve.dp, port)
        if stack_changes or valve.stale_root:
            self.logger.info(f'{stack_changes} stack ports changed state, stale root {valve.stale_root}')
            valve.stale_root = False
            notify_dps = {}
            for stack_valve in stacked_valves:
                if not stack_valve.dp.dyn_running:
                    continue
                ofmsgs_by_valve[stack_valve].extend(
                    stack_valve.add_vlans(stack_valve.dp.vlans.values()))
                for port in stack_valve.dp.stack_ports():
                    ofmsgs_by_valve[stack_valve].extend(
                        stack_valve.switch_manager.del_port(port))
                ofmsgs_by_valve[stack_valve].extend(
                    stack_valve.stack_manager.add_tunnel_acls())
                path_port = stack_valve.stack_manager.chosen_towards_port
                path_port_number = path_port.number if path_port else 0.0
                self._set_var(
                    'dp_root_hop_port', path_port_number, labels=stack_valve.dp.base_prom_labels())
                notify_dps.setdefault(stack_valve.dp.name, {})['root_hop_port'] = path_port_number
            # Find the first valve with a valid stack and trigger notification.
            for stack_valve in stacked_valves:
                if stack_valve.dp.stack.graph:
                    self.notify(
                        {'STACK_TOPO_CHANGE': {
                            'stack_root': stack_valve.dp.stack.root_name,
                            'graph': stack_valve.dp.stack.get_node_link_data(),
                            'dps': notify_dps
                        }})
                    break
        return ofmsgs_by_valve



#!/usr/bin/env python3

"""Implement configuration file parsing."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import re

from faucet import config_parser_util
from faucet.acl import ACL
from faucet.conf import test_config_condition, InvalidConfigError
from faucet.dp import DP
from faucet.meter import Meter
from faucet.port import Port
from faucet.router import Router
from faucet.vlan import VLAN
from faucet.watcher_conf import WatcherConf


V2_TOP_CONFS = (
    'acls',
    'dps',
    'meters',
    'routers',
    'vlans')


def dp_parser(config_file, logname, meta_dp_state=None):
    """Parse a config file into DP configuration objects with hashes of config include/files."""
    conf, _ = config_parser_util.read_config(config_file, logname)
    config_hashes = None
    dps = None

    test_config_condition(conf is None, 'Config file is empty')
    test_config_condition(
        not isinstance(conf, dict),
        'Config file does not have valid syntax')
    version = conf.pop('version', 2)
    test_config_condition(version != 2, 'Only config version 2 is supported')
    config_hashes, config_contents, dps, top_conf = _config_parser_v2(
        config_file, logname, meta_dp_state)
    test_config_condition(dps is None, 'no DPs are not defined')

    return config_hashes, config_contents, dps, top_conf


def _get_vlan_by_key(dp_id, vlan_key, vlans):
    try:
        if vlan_key in vlans:
            return vlans[vlan_key]
    except TypeError as err:
        raise InvalidConfigError(err) from err
    for vlan in vlans.values():
        if vlan_key == vlan.vid:
            return vlan
    test_config_condition(not isinstance(vlan_key, int), (
        f'Implicitly created VLAN {vlan_key} must be an int (not {type(vlan_key)})'))
    # Create VLAN with VID, if not defined.
    return vlans.setdefault(vlan_key, VLAN(vlan_key, dp_id))


def _dp_parse_port(dp_id, port_key, port_conf, vlans):

    def _dp_parse_native_port_vlan():
        if port.native_vlan is not None:
            vlan = _get_vlan_by_key(dp_id, port.native_vlan, vlans)
            port.native_vlan = vlan

    def _dp_parse_tagged_port_vlans():
        if port.tagged_vlans:
            port_tagged_vlans = [
                _get_vlan_by_key(dp_id, vlan_key, vlans) for vlan_key in port.tagged_vlans]
            port.tagged_vlans = port_tagged_vlans

    port = Port(port_key, dp_id, port_conf)
    test_config_condition(str(port_key) not in (str(port.number), port.name), (
        f'Port key {port_key} match port name or port number'))
    _dp_parse_native_port_vlan()
    _dp_parse_tagged_port_vlans()
    return port


def _dp_add_ports(dp, dp_conf, dp_id, vlans):
    ports_conf = dp_conf.get('interfaces', {})
    port_ranges_conf = dp_conf.get('interface_ranges', {})
    # as users can config port VLAN by using VLAN name, we store vid in
    # Port instance instead of VLAN name for data consistency
    test_config_condition(not isinstance(ports_conf, dict), (
        'Invalid syntax in interface config'))
    test_config_condition(not isinstance(port_ranges_conf, dict), (
        'Invalid syntax in interface ranges config'))

    def _map_port_num_to_port(ports_conf):
        port_num_to_port_conf = {}
        for port_key, port_conf in ports_conf.items():
            test_config_condition(not isinstance(port_conf, dict), 'Invalid syntax in port config')
            port_num = port_conf.get('number', port_key)
            try:
                port_num_to_port_conf[port_num] = (port_key, port_conf)
            except TypeError as type_error:
                raise InvalidConfigError('Invalid syntax in port config') from type_error
        return port_num_to_port_conf

    def _parse_port_ranges(port_ranges_conf, port_num_to_port_conf):
        all_port_nums = set()
        for port_range, port_conf in port_ranges_conf.items():
            # port range format: 1-6 OR 1-6,8-9 OR 1-3,5,7-9
            test_config_condition(not isinstance(port_conf, dict), 'Invalid syntax in port config')
            port_nums = set()
            if 'number' in port_conf:
                del port_conf['number']
            for range_ in re.findall(r'(\d+-\d+)', str(port_range)):
                start_num, end_num = [int(num) for num in range_.split('-')]
                test_config_condition(start_num >= end_num, (
                    f'Incorrect port range ({start_num} - {end_num})'))
                port_nums.update(range(start_num, end_num + 1))
                port_range = re.sub(range_, '', port_range)
            other_nums = [int(p) for p in re.findall(r'\d+', str(port_range))]
            port_nums.update(other_nums)
            test_config_condition(
                not port_nums, 'interface-ranges contain invalid config')
            test_config_condition(
                port_nums.intersection(all_port_nums), 'interfaces-ranges cannot overlap')
            all_port_nums.update(port_nums)
            for port_num in port_nums:
                if port_num in port_num_to_port_conf:
                    # port range config has lower priority than individual port config
                    for attr, value in port_conf.items():
                        port_num_to_port_conf[port_num][1].setdefault(attr, value)
                else:
                    port_num_to_port_conf[port_num] = (port_num, port_conf)

    port_num_to_port_conf = _map_port_num_to_port(ports_conf)
    _parse_port_ranges(port_ranges_conf, port_num_to_port_conf)

    for port_num, port_conf in port_num_to_port_conf.values():
        port = _dp_parse_port(dp_id, port_num, port_conf, vlans)
        dp.add_port(port)


def _parse_acls(dp, acls_conf):
    for acl_key, acl_conf in acls_conf.items():
        acl = ACL(acl_key, dp.dp_id, acl_conf)
        dp.add_acl(acl_key, acl)


def _parse_routers(dp, routers_conf):
    for router_key, router_conf in routers_conf.items():
        router = Router(router_key, dp.dp_id, router_conf)
        dp.add_router(router_key, router)


def _parse_meters(dp, meters_conf):
    for meter_key, meter_conf in meters_conf.items():
        meter = Meter(meter_key, dp.dp_id, meter_conf)
        dp.meters[meter_key] = meter


def _parse_dp(dp_key, dp_conf, acls_conf, meters_conf, routers_conf, vlans_conf):
    test_config_condition(not isinstance(dp_conf, dict), 'DP config must be dict')
    dp = DP(dp_key, dp_conf.get('dp_id', None), dp_conf)
    test_config_condition(dp.name != dp_key, (
        f'DP key {dp_key} and DP name must match'))
    vlans = {}
    vids = set()
    for vlan_key, vlan_conf in vlans_conf.items():
        vlan = VLAN(vlan_key, dp.dp_id, vlan_conf)
        test_config_condition(str(vlan_key) not in (str(vlan.vid), vlan.name), (
            f'VLAN {vlan_key} key must match VLAN name or VLAN VID'))
        test_config_condition(not isinstance(vlan_key, (str, int)), (
            f'VLAN {vlan_key} key must not be type {type(vlan_key)}'))
        test_config_condition(vlan.vid in vids, (
            f'VLAN VID {vlan.vid} multiply configured'))
        vlans[vlan_key] = vlan
        vids.add(vlan.vid)
    _parse_acls(dp, acls_conf)
    _parse_routers(dp, routers_conf)
    _parse_meters(dp, meters_conf)
    _dp_add_ports(dp, dp_conf, dp.dp_id, vlans)
    return (dp, vlans)


def _dp_parser_v2(dps_conf, acls_conf, meters_conf,
                  routers_conf, vlans_conf, meta_dp_state):
    # pylint: disable=invalid-name
    dp_vlans = []
    for dp_key, dp_conf in dps_conf.items():
        try:
            dp, vlans = _parse_dp(
                dp_key, dp_conf, acls_conf, meters_conf, routers_conf, vlans_conf)
            dp_vlans.append((dp, vlans))
        except InvalidConfigError as err:
            raise InvalidConfigError(f'DP {dp_key}: {err}') from err

    # Some VLANs are created implicitly just by referencing them in tagged/native,
    # so we must make them available to all DPs.
    implicit_vids = set()
    for dp, vlans in dp_vlans:
        implicit_vids.update(set(vlans.keys()) - set(vlans_conf.keys()))
    dps = []
    for dp, vlans in dp_vlans:
        for vlan_key in implicit_vids:
            if vlan_key not in vlans:
                vlans[vlan_key] = VLAN(vlan_key, dp.dp_id)
        dp.reset_refs(vlans=vlans)
        dps.append(dp)

    for dp in dps:
        dp.finalize_config(dps)
    for dp in dps:
        dp.resolve_stack_topology(dps, meta_dp_state)
    for dp in dps:
        dp.finalize()

    dpid_refs = set()
    for dp in dps:
        test_config_condition(dp.dp_id in dpid_refs, (
            f'DPID {dp.dp_id} is duplicated'))
        dpid_refs.add(dp.dp_id)

    routers_referenced = set()
    for dp in dps:
        routers_referenced.update(dp.routers.keys())
    for router in routers_conf:
        test_config_condition(router not in routers_referenced, (
            f'router {router} configured but not used by any DP'))

    return dps


def dp_preparsed_parser(top_confs, meta_dp_state):
    """Parse a preparsed (after include files have been applied) FAUCET config."""
    local_top_confs = copy.deepcopy(top_confs)
    return _dp_parser_v2(
        local_top_confs.get('dps', {}),
        local_top_confs.get('acls', {}),
        local_top_confs.get('meters', {}),
        local_top_confs.get('routers', {}),
        local_top_confs.get('vlans', {}),
        meta_dp_state)


def _config_parser_v2(config_file, logname, meta_dp_state):
    config_path = config_parser_util.dp_config_path(config_file)
    top_confs = {top_conf: {} for top_conf in V2_TOP_CONFS}
    config_hashes = {}
    config_contents = {}
    dps = None

    if not config_parser_util.dp_include(
            config_hashes, config_contents, config_path, logname, top_confs):
        raise InvalidConfigError(f'Error found while loading config file: {config_path}')

    if not top_confs['dps']:
        raise InvalidConfigError(f'DPs not configured in file: {config_path}')

    dps = dp_preparsed_parser(top_confs, meta_dp_state)
    return (config_hashes, config_contents, dps, top_confs)


def watcher_parser(config_file, logname, prom_client):
    """Return Watcher instances from config."""
    conf, _ = config_parser_util.read_config(config_file, logname)
    conf_hash = config_parser_util.config_file_hash(config_file)
    faucet_config_files, faucet_conf_hashes, result = _watcher_parser_v2(
        conf, logname, prom_client)
    return conf_hash, faucet_config_files, faucet_conf_hashes, result


def _parse_dps_for_watchers(conf, logname, meta_dp_state=None):
    all_dps_list = []
    faucet_conf_hashes = {}

    if not isinstance(conf, dict):
        raise InvalidConfigError('Gauge config not valid')

    faucet_config_files = conf.get('faucet_configs', [])
    for faucet_config_file in faucet_config_files:
        conf_hashes, _, dp_list, _ = dp_parser(faucet_config_file, logname)
        if dp_list:
            faucet_conf_hashes[faucet_config_file] = conf_hashes
            all_dps_list.extend(dp_list)

    faucet_config = conf.get('faucet', None)
    if faucet_config:
        all_dps_list.extend(dp_preparsed_parser(faucet_config, meta_dp_state))

    dps = {dp.name: dp for dp in all_dps_list}
    if not dps:
        raise InvalidConfigError(
            'Gauge configured without any FAUCET configuration')
    return faucet_config_files, faucet_conf_hashes, dps


def _watcher_parser_v2(conf, logname, prom_client):
    logger = config_parser_util.get_logger(logname)

    if conf is None:
        conf = {}
    faucet_config_files, faucet_conf_hashes, dps = _parse_dps_for_watchers(
        conf, logname)
    dbs = conf.pop('dbs')

    result = []
    for watcher_name, watcher_conf in conf['watchers'].items():
        if watcher_conf.get('all_dps', False):
            watcher_dps = dps.keys()
        else:
            watcher_dps = watcher_conf['dps']
        # Watcher config has a list of DPs, but actually a WatcherConf is
        # created for each DP.
        # TODO: refactor watcher_conf as a container.
        for dp_name in watcher_dps:
            if dp_name not in dps:
                logger.error(f'DP {dp_name} in Gauge but not configured in FAUCET')
                continue
            dp = dps[dp_name]
            if 'dbs' in watcher_conf:
                watcher_dbs = watcher_conf['dbs']
            elif 'db' in watcher_conf:
                watcher_dbs = [watcher_conf['db']]
            else:
                raise InvalidConfigError('Watcher configured without DB')
            for db in watcher_dbs:
                watcher = WatcherConf(watcher_name, dp.dp_id, watcher_conf, prom_client)
                watcher.add_db(dbs[db])
                watcher.add_dp(dp)
                result.append(watcher)

    return faucet_config_files, faucet_conf_hashes, result


def get_config_for_api(valves):
    """Return config as dict for all DPs."""
    config = {i: {} for i in V2_TOP_CONFS}
    for valve in valves.values():
        valve_conf = valve.get_config_dict()
        for i in V2_TOP_CONFS:
            if i in valve_conf:
                config[i].update(valve_conf[i])  # pytype: disable=attribute-error
    return config



#!/usr/bin/env python3

"""Run AFL repeatedly with externally supplied generated packet from STDIN."""

import logging
import sys

import afl
from os_ken.controller import dpset
from faucet import faucet
import fake_packet

ROUNDS = 1
logging.disable(logging.CRITICAL)


def main():
    """Run AFL repeatedly with externally supplied generated packet from STDIN."""
    application = faucet.Faucet(dpset=dpset.DPSet())
    application.start()

    # make sure dps are running
    if application.valves_manager is not None:
        for valve in list(application.valves_manager.valves.values()):
            state = valve.dp.dyn_finalized
            valve.dp.dyn_finalized = False
            valve.dp.running = True
            valve.dp.dyn_finalized = state

    while afl.loop(ROUNDS):  # pylint: disable=c-extension-no-member
        # receive input from afl
        rcv = sys.stdin.read()
        data = None
        try:
            data = bytearray.fromhex(rcv)  # pytype: disable=missing-parameter,wrong-arg-types
        except (ValueError, TypeError):
            continue

        # create fake packet
        _dp = fake_packet.Datapath(1)
        msg = fake_packet.Message(datapath=_dp, cookie=15243729, port=1, data=data, in_port=1)
        pkt = fake_packet.RyuEvent(msg)

        # send fake packet to faucet
        application.packet_in_handler(pkt)


if __name__ == "__main__":
    main()



"""Prometheus for Gauge."""

# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from functools import partial
from prometheus_client import Gauge

from faucet.gauge_pollers import GaugePortStatsPoller, GaugePortStatePoller, GaugeFlowTablePoller
from faucet.prom_client import PromClient


PROM_PREFIX_DELIM = '_'
PROM_PORT_PREFIX = 'of_port'
PROM_PORT_STATE_VARS = (
    'reason',
    'state',
    'curr_speed',
    'max_speed',
)
PROM_PORT_VARS = (
    'tx_packets',
    'rx_packets',
    'tx_bytes',
    'rx_bytes',
    'tx_dropped',
    'rx_dropped',
    'tx_errors',
    'rx_errors')
PROM_FLOW_VARS = (
    'flow_byte_count',
    'flow_packet_count'
)
PROM_METER_PREFIX = 'of_meter'
PROM_METER_VARS = (
    'flow_count',
    'byte_in_count',
    'packet_in_count',
    'byte_band_count',
    'packet_band_count'
)


class GaugePrometheusClient(PromClient):
    """Wrapper for Prometheus client that is shared between all pollers."""

    def __init__(self, reg=None):
        super().__init__(reg=reg)
        self.table_tags = collections.defaultdict(set)
        self.metrics = {}
        self.dp_status = Gauge(
            'dp_status',
            'status of datapaths',
            self.REQUIRED_LABELS,
            registry=self._reg)
        self.reregister_nonflow_vars()

    def _reregister_var(self, var_key, var_func):
        try:
            self._reg.unregister(self.metrics[var_key])
        except KeyError:
            pass
        self.metrics[var_key] = var_func()

    def reregister_nonflow_vars(self):
        """Reset all metrics to empty."""
        for prom_var in PROM_PORT_VARS + PROM_PORT_STATE_VARS:
            exported_prom_var = PROM_PREFIX_DELIM.join(
                (PROM_PORT_PREFIX, prom_var))
            self._reregister_var(
                exported_prom_var,
                partial(
                    Gauge,
                    exported_prom_var,
                    '',
                    self.REQUIRED_LABELS + ['port', 'port_description'],
                    registry=self._reg))
        for prom_var in PROM_METER_VARS:
            exported_prom_var = PROM_PREFIX_DELIM.join(
                (PROM_METER_PREFIX, prom_var))
            self._reregister_var(
                exported_prom_var,
                partial(
                    Gauge,
                    exported_prom_var,
                    '',
                    self.REQUIRED_LABELS + ['meter_id'],
                    registry=self._reg))

    def reregister_flow_vars(self, table_name, table_tags):
        """Register the flow variables needed for this client"""
        for prom_var in PROM_FLOW_VARS:
            table_prom_var = PROM_PREFIX_DELIM.join((prom_var, table_name))
            self._reregister_var(
                table_prom_var,
                partial(
                    Gauge,
                    table_prom_var,
                    '',
                    list(table_tags),
                    registry=self._reg))


class GaugePortStatsPrometheusPoller(GaugePortStatsPoller):
    """Exports port stats to Prometheus."""

    def __init__(self, conf, logger, prom_client):
        super().__init__(
            conf, logger, prom_client)
        self.prom_client.start(
            self.conf.prometheus_port, self.conf.prometheus_addr, self.conf.prometheus_test_thread)

    def _format_stat_pairs(self, delim, stat):
        stat_pairs = (
            ((delim.join((PROM_PORT_PREFIX, prom_var)),), getattr(stat, prom_var))
            for prom_var in PROM_PORT_VARS)
        return self._format_stats(delim, stat_pairs)

    def _update(self, rcv_time, msg):
        for stat in msg.body:
            port_labels = self.dp.port_labels(stat.port_no)
            for stat_name, stat_val in self._format_stat_pairs(
                    PROM_PREFIX_DELIM, stat):
                self.prom_client.metrics[stat_name].labels(**port_labels).set(stat_val)


class GaugeMeterStatsPrometheusPoller(GaugePortStatsPoller):
    """Exports meter stats to Prometheus."""

    def __init__(self, conf, logger, prom_client):
        super().__init__(
            conf, logger, prom_client)
        self.prom_client.start(
            self.conf.prometheus_port, self.conf.prometheus_addr, self.conf.prometheus_test_thread)

    def _format_stat_pairs(self, delim, stat):
        band_stats = stat.band_stats[0]
        stat_pairs = (
            (('flow', 'count'), stat.flow_count),
            (('byte', 'in', 'count'), stat.byte_in_count),
            (('packet', 'in', 'count'), stat.packet_in_count),
            (('byte', 'band', 'count'), band_stats.byte_band_count),
            (('packet', 'band', 'count'), band_stats.packet_band_count),
        )
        return self._format_stats(delim, stat_pairs)

    def _update(self, rcv_time, msg):
        for stat in msg.body:
            meter_labels = self.dp.base_prom_labels()
            meter_labels.update({'meter_id': stat.meter_id})
            for stat_name, stat_val in self._format_stat_pairs(
                    PROM_PREFIX_DELIM, stat):
                stat_name = PROM_PREFIX_DELIM.join((PROM_METER_PREFIX, stat_name))
                self.prom_client.metrics[stat_name].labels(**meter_labels).set(stat_val)


class GaugePortStatePrometheusPoller(GaugePortStatePoller):
    """Export port state changes to Prometheus."""

    def _update(self, rcv_time, msg):
        port_no = msg.desc.port_no
        port = self.dp.ports.get(port_no, None)
        if port is None:
            return
        port_labels = self.dp.port_labels(port_no)
        for prom_var in PROM_PORT_STATE_VARS:
            exported_prom_var = PROM_PREFIX_DELIM.join((PROM_PORT_PREFIX, prom_var))
            msg_value = msg.reason if prom_var == 'reason' else getattr(msg.desc, prom_var)
            self.prom_client.metrics[exported_prom_var].labels(**port_labels).set(msg_value)

    def send_req(self):
        """Send a stats request to a datapath."""
        raise NotImplementedError  # pragma: no cover

    def no_response(self):
        """Called when a polling cycle passes without receiving a response."""
        raise NotImplementedError  # pragma: no cover


class GaugeFlowTablePrometheusPoller(GaugeFlowTablePoller):
    """Export flow table entries to Prometheus."""

    def _update(self, rcv_time, msg):
        jsondict = msg.to_jsondict()
        for stats_reply in jsondict['OFPFlowStatsReply']['body']:
            stats = stats_reply['OFPFlowStats']
            # TODO: labels based on matches will be dynamic
            # Work around this by unregistering/registering the entire variable.
            for var, tags, count in self._parse_flow_stats(stats):
                table_id = int(tags['table_id'])
                table_name = self.dp.table_by_id(table_id).name
                table_tags = self.prom_client.table_tags[table_name]
                tags_keys = set(tags.keys())
                if tags_keys != table_tags:
                    unreg_tags = tags_keys - table_tags
                    if unreg_tags:
                        table_tags.update(unreg_tags)
                        self.prom_client.reregister_flow_vars(
                            table_name, table_tags)
                        self.logger.info(  # pylint: disable=logging-not-lazy
                            'Adding tags %s to %s for table %s' % (
                                unreg_tags, table_tags, table_name))
                    # Add blank tags for any tags not present.
                    missing_tags = table_tags - tags_keys
                    for tag in missing_tags:
                        tags[tag] = ''
                table_prom_var = PROM_PREFIX_DELIM.join((var, table_name))
                try:
                    self.prom_client.metrics[table_prom_var].labels(**tags).set(count)
                except ValueError:
                    self.logger.error(  # pylint: disable=logging-not-lazy
                        'labels %s versus %s incorrect on %s' % (
                            tags, table_tags, table_prom_var))



#!/usr/bin/env python3

"""Launch forwarder script for Faucet/Gauge"""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import sys

from pbr.version import VersionInfo

if sys.version_info < (3,) or sys.version_info < (3, 6):
    raise ImportError("""You are trying to run faucet on python {py}

Faucet is not compatible with python {py}, please upgrade to python 3.6 or newer."""
                      .format(py='.'.join([str(v) for v in sys.version_info[:3]])))

RYU_OPTIONAL_ARGS = [
    ('ca-certs', 'CA certificates'),
    ('config-dir', """Path to a config directory to pull `*.conf` files
                      from. This file set is sorted, so as to provide a
                      predictable parse order if individual options are
                      over-ridden. The set is parsed after the file(s)
                      specified via previous --config-file, arguments hence
                      over-ridden options in the directory take precedence."""),
    ('config-file', """Path to a config file to use. Multiple config files
                       can be specified, with values in later files taking
                       precedence. Defaults to None.""", "/etc/faucet/ryu.conf"),
    ('ctl-cert', 'controller certificate'),
    ('ctl-privkey', 'controller private key'),
    ('default-log-level', 'default log level'),
    ('log-config-file', 'Path to a logging config file to use'),
    ('log-dir', 'log file directory'),
    ('log-file', 'log file name'),
    ('log-file-mode', 'default log file permission'),
    ('observe-links', 'observe link discovery events'),
    ('ofp-listen-host', 'openflow listen host (default 0.0.0.0)'),
    ('ofp-ssl-listen-port', 'openflow ssl listen port (default: 6653)'),
    ('ofp-switch-address-list', """list of IP address and port pairs (default empty).
                                   e.g., "127.0.0.1:6653,[::1]:6653"""),
    ('ofp-switch-connect-interval', 'interval in seconds to connect to switches (default 1)'),
    ('ofp-tcp-listen-port', 'openflow tcp listen port (default: 6653)'),
    ('pid-file', 'pid file name'),
    ('user-flags', 'Additional flags file for user applications'),
]


def parse_args(sys_args):
    """Parse Faucet/Gauge arguments.

    Returns:
        argparse.Namespace: command line arguments
    """

    args = argparse.ArgumentParser(
        prog='faucet', description='Faucet SDN Controller')
    args.add_argument('--gauge', action='store_true', help='run Gauge instead')
    args.add_argument(
        '-v', '--verbose', action='store_true', help='produce verbose output')
    args.add_argument(
        '-V', '--version', action='store_true', help='print version and exit')
    args.add_argument(
        '--use-stderr', action='store_true', help='log to standard error')
    args.add_argument(
        '--use-syslog', action='store_true', help='output to syslog')
    args.add_argument(
        '--ryu-app-lists',
        action='append',
        help='add Ryu app (can be specified multiple times)',
        metavar='APP')

    for ryu_arg in RYU_OPTIONAL_ARGS:
        if len(ryu_arg) >= 3:
            args.add_argument(
                f'--ryu-{ryu_arg[0]}',
                help=ryu_arg[1],
                default=ryu_arg[2])
        else:
            args.add_argument(
                f'--ryu-{ryu_arg[0]}',
                help=ryu_arg[1])

    return args.parse_args(sys_args)


def print_version():
    """Print version number and exit."""
    version = VersionInfo('c65faucet').semantic_version().release_string()
    message = 'c65faucet %s' % version
    print(message)


def build_ryu_args(argv):
    args = parse_args(argv[1:])

    # Checking version number?
    if args.version:
        print_version()
        return []

    prog = os.path.basename(argv[0])
    ryu_args = []

    # Handle log location
    if args.use_stderr:
        ryu_args.append('--use-stderr')
    if args.use_syslog:
        ryu_args.append('--use-syslog')

    # Verbose output?
    if args.verbose:
        ryu_args.append('--verbose')

    for arg, val in vars(args).items():
        if not val or not arg.startswith('ryu'):
            continue
        if arg == 'ryu_app_lists':
            continue
        if arg == 'ryu_config_file' and not os.path.isfile(val):
            continue
        arg_name = arg.replace('ryu_', '').replace('_', '-')
        ryu_args.append(f'--{arg_name}={val}')

    # Running Faucet or Gauge?
    if args.gauge or os.path.basename(prog) == 'gauge':
        ryu_args.append('faucet.gauge')
    else:
        ryu_args.append('faucet.faucet')

    # Check for additional Ryu apps.
    if args.ryu_app_lists:
        ryu_args.extend(args.ryu_app_lists)

    # Replace current process with ryu-manager from PATH (no PID change).
    ryu_args.insert(0, 'osken-manager')
    return ryu_args


def main():
    """Main program."""
    ryu_args = build_ryu_args(sys.argv)
    if ryu_args:
        os.execvp(ryu_args[0], ryu_args)


if __name__ == '__main__':
    main()



"""Compose ACLs on ports."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from faucet import valve_of
from faucet import valve_packet
from faucet.valve_manager_base import ValveManagerBase
from faucet.conf import InvalidConfigError


def push_vlan(acl_table, vlan_vid):
    """Push a VLAN tag with optional selection of eth type."""
    vid = vlan_vid
    vlan_eth_type = None
    if isinstance(vlan_vid, dict):
        vid = vlan_vid['vid']
        if 'eth_type' in vlan_vid:
            vlan_eth_type = vlan_vid['eth_type']
    if vlan_eth_type is None:
        return valve_of.push_vlan_act(acl_table, vid)
    return valve_of.push_vlan_act(
        acl_table, vid, eth_type=vlan_eth_type)


def build_ordered_output_actions(acl_table, output_list, tunnel_rules=None, source_id=None):
    """Build actions from ordered ACL output list"""
    output_actions = []
    output_ports = []
    output_ofmsgs = []
    output_inst = []
    for action in output_list:
        for key, value in action.items():
            if key == 'pop_vlans':
                for _ in range(value):
                    output_actions.append(valve_of.pop_vlan())
            if key == 'vlan_vid':
                output_actions.extend(push_vlan(acl_table, value))
            if key == 'swap_vid':
                output_actions.append(acl_table.set_vlan_vid(value))
            if key == 'vlan_vids':
                for vlan_vid in value:
                    output_actions.extend(push_vlan(acl_table, vlan_vid))
            if key == 'set_fields':
                for set_field in value:
                    output_actions.append(acl_table.set_field(**set_field))
            if key == 'port':
                output_ports.append(value)
                output_actions.append(valve_of.output_port(value))
            if key == 'ports':
                for output_port in value:
                    output_ports.append(output_port)
                    output_actions.append(valve_of.output_port(output_port))
            if key == 'failover':
                group_id = value['group_id']
                buckets = []
                for port in value['ports']:
                    buckets.append(valve_of.bucket(
                        watch_port=port, actions=[valve_of.output_port(port)]))
                output_ofmsgs.extend(valve_of.groupadd_ff(group_id=group_id, buckets=buckets))
                output_actions.append(valve_of.group_act(group_id=group_id))
            if key == 'tunnel' and tunnel_rules and source_id is not None:
                source_rule = tunnel_rules[value][source_id]
                _, tunnel_actions, tunnel_ofmsgs, tunnel_inst = build_output_actions(acl_table,
                                                                                     source_rule)
                output_actions.extend(tunnel_actions)
                output_ofmsgs.extend(tunnel_ofmsgs)
                output_inst.extend(tunnel_inst)
            if key == 'goto':
                output_inst.append(valve_of.goto_table_id(value))
    return (output_ports, output_actions, output_ofmsgs, output_inst)


def rewrite_vlan(acl_table, output_dict):
    """Implement actions to rewrite VLAN headers."""
    vlan_actions = []
    if 'pop_vlans' in output_dict:
        for _ in range(output_dict['pop_vlans']):
            vlan_actions.append(valve_of.pop_vlan())
    # if vlan tag is specified, push it.
    if 'vlan_vid' in output_dict:
        vlan_actions.extend(push_vlan(acl_table, output_dict['vlan_vid']))
    # swap existing VID
    elif 'swap_vid' in output_dict:
        vlan_actions.append(
            acl_table.set_vlan_vid(output_dict['swap_vid']))
    # or, if a list, push them all (all with type Q).
    elif 'vlan_vids' in output_dict:
        for vlan_vid in output_dict['vlan_vids']:
            vlan_actions.extend(push_vlan(acl_table, vlan_vid))
    return vlan_actions


def build_ct_actions(acl_table, ct_dict):
    """Build conntrack action from ACL rule"""

    if 'clear' in ct_dict and ct_dict['clear']:
        return valve_of.ct_clear()

    ct_actions = []

    if 'nat' in ct_dict:
        nat_args = {
            'flags': ct_dict['nat'].get('flags', 0),
            'range_ipv4_min': ct_dict['nat'].get('range_ipv4_min', ''),
            'range_ipv4_max': ct_dict['nat'].get('range_ipv4_max', ''),
            'range_ipv6_min': ct_dict['nat'].get('range_ipv6_min', ''),
            'range_ipv6_max': ct_dict['nat'].get('range_ipv6_max', ''),
            'range_proto_min': ct_dict['nat'].get('range_proto_min', None),
            'range_proto_max': ct_dict['nat'].get('range_proto_max', None),
        }
        ct_actions.append(valve_of.ct_nat(**nat_args))

    ct_args = {
        'flags': ct_dict.get('flags', 0),
        'actions': ct_actions,
        'alg': ct_dict.get('alg', 0),
        'recirc_table': ct_dict.get('table'),
        'zone_ofs_nbits': ct_dict.get('zone'),
        'zone_src': ct_dict.get('zone_src', None),
    }

    return valve_of.ct(**ct_args)


def build_output_actions(acl_table, output_dict, tunnel_rules=None, source_id=None):
    """Implement actions to alter packet/output."""
    if isinstance(output_dict, (list, tuple)):
        return build_ordered_output_actions(acl_table, output_dict, tunnel_rules, source_id)
    output_actions = []
    output_inst = []
    output_port = None
    ofmsgs = []
    # rewrite any VLAN headers first always
    vlan_actions = rewrite_vlan(acl_table, output_dict)
    if vlan_actions:
        output_actions.extend(vlan_actions)
    if 'set_fields' in output_dict:
        for set_field in output_dict['set_fields']:
            output_actions.append(acl_table.set_field(**set_field))
    if 'port' in output_dict:
        output_port = output_dict['port']
        output_actions.append(valve_of.output_port(output_port))
    if 'ports' in output_dict:
        for output_port in output_dict['ports']:
            output_actions.append(valve_of.output_port(output_port))
    if 'failover' in output_dict:
        failover = output_dict['failover']
        group_id = failover['group_id']
        buckets = []
        for port in failover['ports']:
            buckets.append(valve_of.bucket(
                watch_port=port, actions=[valve_of.output_port(port)]))
        ofmsgs.extend(valve_of.groupadd_ff(group_id=group_id, buckets=buckets))
        output_actions.append(valve_of.group_act(group_id=group_id))
    if 'tunnel' in output_dict and tunnel_rules and source_id is not None:
        tunnel_id = output_dict['tunnel']
        source_rule = tunnel_rules[tunnel_id][source_id]
        _, tunnel_actions, tunnel_ofmsgs, tunnel_inst = build_output_actions(acl_table, source_rule)
        output_actions.extend(tunnel_actions)
        tunnel_ofmsgs.extend(tunnel_ofmsgs)
        output_inst.extend(tunnel_inst)
    if 'goto' in output_dict:
        output_inst.append(valve_of.goto_table_id(output_dict['goto']))
    return (output_port, output_actions, ofmsgs, output_inst)


def build_acl_entry(  # pylint: disable=too-many-arguments,too-many-branches,too-many-statements
        acl_table, rule_conf, meters,
        acl_allow_inst, acl_force_port_vlan_inst,
        port_num=None, vlan_vid=None, tunnel_rules=None, source_id=None):
    """Build flow/groupmods for one ACL rule entry."""
    acl_inst = []
    acl_act = []
    acl_match_dict = {}
    acl_ofmsgs = []
    acl_cookie = None
    allow_inst = acl_allow_inst

    for attrib, attrib_value in rule_conf.items():
        # if attrib == 'in_port':
        #     continue
        if attrib == 'cookie':
            acl_cookie = attrib_value
            continue
        if attrib == 'description':
            continue
        if attrib == 'actions':
            allow = False
            allow_specified = False
            if 'allow' in attrib_value:
                allow_specified = True
                if attrib_value['allow'] == 1:
                    allow = True
            if 'force_port_vlan' in attrib_value:
                if attrib_value['force_port_vlan'] == 1:
                    allow_inst = acl_force_port_vlan_inst
            if 'meter' in attrib_value:
                meter_name = attrib_value['meter']
                acl_inst.append(valve_of.apply_meter(meters[meter_name].meter_id))
            if 'mirror' in attrib_value:
                port_no = attrib_value['mirror']
                acl_act.append(valve_of.output_port(port_no))
                if not allow_specified:
                    allow = True
            if 'output' in attrib_value:
                output_port, output_actions, output_ofmsgs, output_inst = build_output_actions(
                    acl_table, attrib_value['output'], tunnel_rules, source_id)
                acl_act.extend(output_actions)
                acl_ofmsgs.extend(output_ofmsgs)
                acl_inst.extend(output_inst)

                # if port specified, output packet now and exit pipeline.
                if not allow and output_port is not None:
                    continue
            if 'ct' in attrib_value:
                ct_action = build_ct_actions(acl_table, attrib_value['ct'])
                acl_act.append(ct_action)

            if allow:
                acl_inst.extend(allow_inst)
        else:
            acl_match_dict[attrib] = attrib_value
    if port_num is not None:
        # This overwrites the `in_port` match if it is specified in the ACL config
        acl_match_dict['in_port'] = port_num
    if vlan_vid is not None:
        # This overwrites the `vlan_vid` match if it is specified in the ACL config
        acl_match_dict['vlan_vid'] = valve_of.vid_present(vlan_vid)
    try:
        acl_match = valve_of.match_from_dict(acl_match_dict)
    except TypeError as type_error:
        raise InvalidConfigError('invalid match type in ACL') from type_error
    if acl_act:
        acl_inst.append(valve_of.apply_actions(acl_act))
    return (acl_match, acl_inst, acl_cookie, acl_ofmsgs)


def build_tunnel_ofmsgs(rule_conf, acl_table, priority,
                        port_num=None, vlan_vid=None, flowdel=False,
                        reverse=False):
    """Build a specific tunnel only ofmsgs"""
    ofmsgs = []
    acl_inst = []
    acl_match = []
    acl_match_dict = {}
    _, output_actions, output_ofmsgs, output_inst = build_output_actions(acl_table, rule_conf)
    ofmsgs.extend(output_ofmsgs)
    acl_inst.extend(output_inst)
    acl_inst.append(valve_of.apply_actions(output_actions))
    if port_num is not None:
        acl_match_dict['in_port'] = port_num
    if vlan_vid is not None:
        acl_match_dict['vlan_vid'] = valve_of.vid_present(vlan_vid)
    if reverse:
        acl_match_dict['vlan_pcp'] = valve_of.PCP_TUNNEL_REVERSE_DIRECTION_FLAG
    else:
        acl_match_dict['vlan_pcp'] = valve_of.PCP_TUNNEL_FLAG
    try:
        acl_match = valve_of.match_from_dict(acl_match_dict)
    except TypeError as type_error:
        raise InvalidConfigError('invalid match type in ACL') from type_error
    flowmod = acl_table.flowmod(acl_match, priority=priority, inst=tuple(acl_inst))
    if flowdel:
        ofmsgs.append(acl_table.flowdel(match=acl_match, priority=priority, strict=False))
    ofmsgs.append(flowmod)
    return ofmsgs


def build_rule_ofmsgs(rule_conf, acl_table,
                      acl_allow_inst, acl_force_port_vlan_inst,
                      highest_priority, acl_rule_priority, meters,
                      exact_match, port_num=None, vlan_vid=None,
                      tunnel_rules=None, source_id=None, flowdel=False):
    """Build an ACL rule and return OFMSGs"""
    ofmsgs = []
    acl_match, acl_inst, acl_cookie, acl_ofmsgs = build_acl_entry(
        acl_table, rule_conf, meters,
        acl_allow_inst, acl_force_port_vlan_inst,
        port_num, vlan_vid, tunnel_rules, source_id)
    ofmsgs.extend(acl_ofmsgs)
    priority = acl_rule_priority
    if exact_match:
        priority = highest_priority
    flowmod = acl_table.flowmod(
        acl_match, priority=priority, inst=tuple(acl_inst), cookie=acl_cookie)
    if flowdel:
        ofmsgs.append(acl_table.flowdel(match=acl_match, priority=priority))
    ofmsgs.append(flowmod)
    return ofmsgs


def build_acl_ofmsgs(acls, acl_table,
                     acl_allow_inst, acl_force_port_vlan_inst,
                     highest_priority, meters,
                     exact_match, port_num=None, vlan_vid=None,
                     tunnel_rules=None, source_id=None, flowdel=False):
    """Build flow/groupmods for all entries in an ACL."""
    ofmsgs = []
    acl_rule_priority = highest_priority
    for acl in acls:
        for rule_conf in acl.rules:
            ofmsgs.extend(build_rule_ofmsgs(
                rule_conf, acl_table,
                acl_allow_inst, acl_force_port_vlan_inst,
                highest_priority, acl_rule_priority, meters,
                exact_match, port_num, vlan_vid,
                tunnel_rules=tunnel_rules, source_id=source_id, flowdel=flowdel))
            acl_rule_priority -= 1
    return ofmsgs


def build_acl_port_of_msgs(acl, vid, port_num, acl_table, goto_table, priority):
    """A Helper function for building Openflow Mod Messages for Port ACLs"""
    ofmsgs = None
    if acl.rules:
        ofmsgs = build_acl_ofmsgs(
            [acl], acl_table,
            [valve_of.goto_table(goto_table)],
            [valve_of.goto_table(goto_table)],
            priority, acl.meter, acl.exact_match,
            vlan_vid=vid, port_num=port_num)
    return ofmsgs


def add_mac_address_to_match(match, eth_src):
    """Add or change the value of a match type"""
    # NOTE: This function has been created to work around for
    # OFPMatch.set_dl_src() not storing persistent changes
    if not eth_src:
        return match

    dict_match = dict(match.items())
    dict_match['eth_src'] = eth_src
    return valve_of.match_from_dict(dict_match)


class ValveAclManager(ValveManagerBase):
    """Handle installation of ACLs on a DP"""

    def __init__(self, port_acl_table, vlan_acl_table, egress_acl_table,
                 pipeline, meters, dp_acls=None):
        self.dp_acls = dp_acls
        self.port_acl_table = port_acl_table
        self.vlan_acl_table = vlan_acl_table
        self.egress_acl_table = egress_acl_table
        self.pipeline = pipeline
        self.acl_priority = self._FILTER_PRIORITY
        self.override_priority = self.acl_priority + 1
        self.auth_priority = self._HIGH_PRIORITY
        self.low_priority = self.auth_priority - 1
        self.meters = meters

    def initialise_tables(self):
        """Install dp acls if configured"""
        ofmsgs = []
        if self.dp_acls:
            acl_allow_inst = self.pipeline.accept_to_vlan()
            acl_force_port_vlan_inst = self.pipeline.accept_to_l2_forwarding()
            ofmsgs.extend(build_acl_ofmsgs(
                self.dp_acls, self.port_acl_table, acl_allow_inst,
                acl_force_port_vlan_inst, self.acl_priority, self.meters,
                False))
        return ofmsgs

    def _port_acls_allowed(self, port):
        return not self.dp_acls and not port.output_only and self.port_acl_table

    def add_port(self, port):
        """Install port acls if configured"""
        ofmsgs = []
        if self._port_acls_allowed(port):
            in_port_match = self.port_acl_table.match(in_port=port.number)
            acl_allow_inst = self.pipeline.accept_to_vlan()
            acl_force_port_vlan_inst = self.pipeline.accept_to_l2_forwarding()
            if port.acls_in:
                ofmsgs.extend(build_acl_ofmsgs(
                    port.acls_in, self.port_acl_table,
                    acl_allow_inst, acl_force_port_vlan_inst,
                    self.acl_priority, self.meters,
                    port.acls_in[0].exact_match, port_num=port.number))
            elif not port.dot1x:
                ofmsgs.append(self.port_acl_table.flowmod(
                    in_port_match,
                    priority=self.acl_priority,
                    inst=tuple(acl_allow_inst)))
        return ofmsgs

    def del_port(self, port):
        ofmsgs = []
        if self._port_acls_allowed(port):
            in_port_match = self.port_acl_table.match(in_port=port.number)
            ofmsgs.append(self.port_acl_table.flowdel(in_port_match, self.acl_priority))
        return ofmsgs

    def cold_start_port(self, port):
        """Reload acl for a port by deleting existing rules and calling
        add_port"""
        ofmsgs = []
        ofmsgs.extend(self.del_port(port))
        ofmsgs.extend(self.add_port(port))
        return ofmsgs

    def del_vlan(self, vlan):
        """Remove VLAN ACLs if configured."""
        ofmsgs = []
        for table in (self.vlan_acl_table, self.egress_acl_table):
            if table is not None:
                ofmsgs.append(table.flowdel(match=table.match(vlan=vlan)))
        return ofmsgs

    def add_vlan(self, vlan, cold_start):
        """Install VLAN ACL if configured."""
        ofmsgs = []
        if vlan.acls_in:
            acl_allow_inst = self.pipeline.accept_to_classification()
            acl_force_port_vlan_inst = self.pipeline.accept_to_l2_forwarding()
            ofmsgs = build_acl_ofmsgs(
                vlan.acls_in, self.vlan_acl_table, acl_allow_inst,
                acl_force_port_vlan_inst, self.acl_priority, self.meters,
                vlan.acls_in[0].exact_match, vlan_vid=vlan.vid)
        if self.egress_acl_table is not None:
            egress_acl_allow_inst = self.pipeline.accept_to_egress()
            if vlan.acls_out:
                ofmsgs.extend(build_acl_ofmsgs(
                    vlan.acls_out, self.egress_acl_table, egress_acl_allow_inst,
                    egress_acl_allow_inst, self.acl_priority, self.meters,
                    vlan.acls_out[0].exact_match, vlan_vid=vlan.vid
                ))
            else:
                ofmsgs.append(self.egress_acl_table.flowmod(
                    self.egress_acl_table.match(vlan=vlan),
                    priority=self.acl_priority,
                    inst=tuple(egress_acl_allow_inst),
                ))
        return ofmsgs

    def add_authed_mac(self, port_num, mac):
        """Add authed mac address"""
        return [self.port_acl_table.flowmod(
            self.port_acl_table.match(in_port=port_num, eth_src=mac),
            priority=self.auth_priority,
            inst=tuple(self.pipeline.accept_to_vlan()))]

    def del_authed_mac(self, port_num, mac=None, strict=True):
        """remove authed mac address"""
        if mac:
            return [self.port_acl_table.flowdel(
                self.port_acl_table.match(in_port=port_num, eth_src=mac),
                priority=self.auth_priority,
                strict=strict)]
        return [self.port_acl_table.flowdel(
            self.port_acl_table.match(in_port=port_num),
            priority=self.auth_priority,
            strict=strict)]

    def del_port_acl(self, acl, port_num, mac=None):
        """Delete ACL rules for Port"""
        def convert_to_flow_del(ofp_flowmods):
            flowdels = []
            for flowmod in ofp_flowmods:
                flowdels.append(self.port_acl_table.flowdel(
                    match=flowmod.match, priority=flowmod.priority))

            return flowdels

        pipeline_vlan_table = self.pipeline.vlan_table
        flowmods = build_acl_port_of_msgs(acl, None, port_num, self.port_acl_table,
                                          pipeline_vlan_table, self.auth_priority)
        for flow in flowmods:
            flow.match = add_mac_address_to_match(flow.match, mac)

        return convert_to_flow_del(flowmods)

    def add_port_acl(self, acl, port_num, mac=None):
        """Create ACL openflow rules for Port"""
        pipeline_vlan_table = self.pipeline.vlan_table
        flowmods = build_acl_port_of_msgs(acl, None, port_num, self.port_acl_table,
                                          pipeline_vlan_table, self.auth_priority)

        for flow in flowmods:
            flow.match = add_mac_address_to_match(flow.match, mac)

        return flowmods

    def create_dot1x_flow_pair(self, port_num, nfv_sw_port_num, mac):
        """Create dot1x flow pair"""
        ofmsgs = [
            self.port_acl_table.flowmod(
                match=self.port_acl_table.match(
                    in_port=port_num,
                    eth_type=valve_packet.ETH_EAPOL),
                priority=self.override_priority,
                inst=(valve_of.apply_actions([
                    self.port_acl_table.set_field(eth_dst=mac),
                    valve_of.output_port(nfv_sw_port_num)]),),
            ),
            self.port_acl_table.flowmod(
                match=self.port_acl_table.match(
                    in_port=nfv_sw_port_num,
                    eth_type=valve_packet.ETH_EAPOL,
                    eth_src=mac),
                priority=self.override_priority,
                inst=(valve_of.apply_actions([
                    self.port_acl_table.set_field(
                        eth_src=valve_packet.EAPOL_ETH_DST),
                    valve_of.output_port(port_num)
                ]),),
            )
        ]

        return ofmsgs

    def del_dot1x_flow_pair(self, port_num, nfv_sw_port_num, mac):
        """Deletes dot1x flow pair"""
        ofmsgs = [
            self.port_acl_table.flowdel(
                match=self.port_acl_table.match(
                    in_port=nfv_sw_port_num,
                    eth_type=valve_packet.ETH_EAPOL,
                    eth_src=mac),
                priority=self.override_priority,
            ),
            self.port_acl_table.flowdel(
                match=self.port_acl_table.match(
                    in_port=port_num,
                    eth_type=valve_packet.ETH_EAPOL),
                priority=self.override_priority,
            )
        ]
        return ofmsgs

    def create_mab_flow(self, port_num, nfv_sw_port_num, mac):
        """
        Create MAB ACL for sending IP Activity to Chewie NFV
            Returns flowmods to send all IP traffic to Chewie

        Args:
            port_num (int): Number of port in
            nfv_sw_port_num(int): Number of port out
            mac(str): MAC address of the valve/port combo

        """
        return self.port_acl_table.flowmod(
            match=self.port_acl_table.match(
                in_port=port_num,
                eth_type=valve_of.ether.ETH_TYPE_IP,
                nw_proto=valve_of.inet.IPPROTO_UDP,
                udp_src=68,
                udp_dst=67,
            ),
            priority=self.low_priority,
            inst=(valve_of.apply_actions([
                self.port_acl_table.set_field(eth_dst=mac),
                valve_of.output_port(nfv_sw_port_num)]),),
        )

    def del_mab_flow(self, port_num, _nfv_sw_port_num, _mac):
        """
        Remove MAB ACL for sending IP Activity to Chewie NFV
            Returns flowmods to send all IP traffic to Chewie

        Args:
            port_num (int): Number of port in
            _nfv_sw_port_num(int): Number of port out
            _mac(str): MAC address of the valve/port combo

        """
        return [self.port_acl_table.flowdel(
            match=self.port_acl_table.match(
                in_port=port_num,
                eth_type=valve_of.ether.ETH_TYPE_IP,
                nw_proto=valve_of.inet.IPPROTO_UDP,
                udp_src=68,
                udp_dst=67,
            ),
            priority=self.low_priority,
            strict=True
        )]

    def build_tunnel_rules_ofmsgs(self, source_id, tunnel_id, acl):
        """Build a tunnel only generated rule"""
        ofmsgs = []
        acl_table = self.port_acl_table
        priority = self.override_priority
        reverse = acl.tunnel_dests[tunnel_id]['reverse']
        ofmsgs.extend(build_tunnel_ofmsgs(
            acl.dyn_tunnel_rules[tunnel_id][source_id], acl_table, priority,
            None, tunnel_id, flowdel=True, reverse=reverse))
        return ofmsgs

    def build_reverse_tunnel_rules_ofmsgs(self, source_id, tunnel_id, acl):
        """Build a (reverse) tunnel only generated rule"""
        ofmsgs = []
        acl_table = self.port_acl_table
        priority = self.override_priority
        if acl.requires_reverse_tunnel(tunnel_id):
            ofmsgs.extend(build_tunnel_ofmsgs(
                acl.dyn_reverse_tunnel_rules[tunnel_id][source_id], acl_table, priority,
                None, tunnel_id, flowdel=True, reverse=True))
        return ofmsgs

    def build_tunnel_acl_rule_ofmsgs(self, source_id, tunnel_id, acl):
        """Build a rule of an ACL that contains a tunnel"""
        ofmsgs = []
        acl_allow_inst = self.pipeline.accept_to_vlan()
        acl_table = self.port_acl_table
        rules = acl.get_tunnel_rules(tunnel_id)
        acl_force_port_vlan_inst = self.pipeline.accept_to_l2_forwarding()
        if self.dp_acls and acl in self.dp_acls:
            ofmsgs.extend(build_acl_ofmsgs(
                self.dp_acls, acl_table, acl_allow_inst,
                acl_force_port_vlan_inst, self.acl_priority, self.meters,
                False, tunnel_rules=acl.dyn_tunnel_rules, source_id=source_id, flowdel=True))
        else:
            for rule_conf in rules:
                rule_index = acl.rules.index(rule_conf)
                priority = self.acl_priority - rule_index
                in_port = acl.tunnel_sources[source_id]['port']
                ofmsgs.extend(build_rule_ofmsgs(
                    rule_conf, acl_table, acl_allow_inst, acl_force_port_vlan_inst,
                    self.acl_priority, priority, self.meters,
                    acl.exact_match, in_port, None, acl.dyn_tunnel_rules, source_id, flowdel=True))
        return ofmsgs

    def add_meters(self, added_meters):
        """Add new meters."""
        ofmsgs = []
        if added_meters:
            for added_meter in added_meters:
                ofmsgs.append(valve_of.meteradd(
                    self.meters.get(added_meter).entry, command=0))
        return ofmsgs

    def del_meters(self, deleted_meters):
        ofmsgs = []
        if deleted_meters:
            deleted_meter_ids = [
                self.meters[meter_key].meter_id for meter_key in deleted_meters]
            ofmsgs.extend([
                valve_of.meterdel(deleted_meter_id) for deleted_meter_id in deleted_meter_ids])
        return ofmsgs



"""Port configuration."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netaddr

from faucet.conf import Conf, test_config_condition
from faucet import valve_of

# Forced port DOWN
STACK_STATE_ADMIN_DOWN = 0
# New instance, initializing stack port
STACK_STATE_INIT = 1
# Incorrect stack configuration
STACK_STATE_BAD = 2
# Stack port up and running, no timeouts, no incorrect cabling
STACK_STATE_UP = 3
# Stack timed out, too many packets lost
STACK_STATE_GONE = 4
# Not stack port configured
STACK_STATE_NONE = -1
STACK_DISPLAY_DICT = {
    STACK_STATE_ADMIN_DOWN: 'ADMIN_DOWN',
    STACK_STATE_INIT: 'INITIALIZING',
    STACK_STATE_BAD: 'BAD',
    STACK_STATE_UP: 'UP',
    STACK_STATE_GONE: 'GONE',
    STACK_STATE_NONE: 'NONE'
}

# LACP not configured
LACP_ACTOR_NOTCONFIGURED = -1
# Not receiving packets from the actor & port is down
LACP_ACTOR_NONE = 0
# Not receiving packets from the actor & port is up
LACP_ACTOR_INIT = 1
# Receiving LACP packets with sync bit set
LACP_ACTOR_UP = 3
# LACP actor is not sending LACP with sync bit set
LACP_ACTOR_NOSYNC = 5
LACP_ACTOR_DISPLAY_DICT = {
    LACP_ACTOR_NOTCONFIGURED: 'NOT_CONFIGURED',
    LACP_ACTOR_NONE: 'NONE',
    LACP_ACTOR_INIT: 'INITIALIZING',
    LACP_ACTOR_UP: 'UP',
    LACP_ACTOR_NOSYNC: 'NO_SYNC'
}

# LACP is not configured
LACP_PORT_NOTCONFIGURED = -1
# Port is not a LACP port on the nominated DP
LACP_PORT_UNSELECTED = 1
# Port is a LACP port on the nominated DP, will send/receive
LACP_PORT_SELECTED = 2
# Port is a LACP port that is in standby
LACP_PORT_STANDBY = 3
LACP_PORT_DISPLAY_DICT = {
    LACP_PORT_NOTCONFIGURED: 'NOT_CONFIGURED',
    LACP_PORT_UNSELECTED: 'UNSELECTED',
    LACP_PORT_SELECTED: 'SELECTED',
    LACP_PORT_STANDBY: 'STANDBY'
}


class Port(Conf):
    """Stores state for ports, including the configuration."""

    defaults = {
        'number': None,
        'name': None,
        'description': None,
        'enabled': True,
        'permanent_learn': False,
        # if True, a host once learned on this port cannot be learned on another port.
        'unicast_flood': True,
        # if True, do classical unicast flooding on this port (False floods ND/ARP/bcast only).
        'mirror': None,
        # If set, mirror packets from that port to this one.
        'native_vlan': None,
        # Set untagged VLAN on this port.
        'tagged_vlans': None,
        # Set tagged VLANs on this port.
        'acl_in': None,
        'acls_in': None,
        # ACL for input on this port.
        'stack': None,
        # Configure a stack peer on this port.
        'max_hosts': 255,
        # maximum number of hosts
        'hairpin': False,
        # if True, then switch unicast and flood between hosts on this port (eg WiFi radio).
        'hairpin_unicast': False,
        # if True, then switch unicast between hosts on this port (eg WiFi radio).
        'lacp': 0,
        # if non 0 (LAG ID), experimental LACP support enabled on this port.
        'lacp_active': False,
        # experimental active LACP
        'lacp_collect_and_distribute': False,
        # if true, forces LACP port to collect and distribute when syncing with the peer.
        'lacp_unselected': False,
        # if true, forces LACP port to be in unselected state
        'lacp_selected': False,
        # if true, forces LACP port to be in the selected state
        'lacp_standby': False,
        # if true, forces LACP port to be in the standby state
        'lacp_passthrough': None,
        # If set, fail the lacp on this port if any of the peer ports are down.
        'lacp_resp_interval': 1,
        # Min time since last LACP response. Used to control rate of response for LACP
        'lacp_port_priority': 255,
        # Sets port priority value sent out in lacp packet
        'lacp_port_id': -1,
        # Sets port id value sent out in lacp packet
        'loop_protect': False,
        # if True, do simple (host/access port) loop protection on this port.
        'loop_protect_external': False,
        # if True, do external (other switch) loop protection on this port.
        'output_only': False,
        # if True, all packets input from this port are dropped.
        'lldp_beacon': {},
        # LLDP beacon configuration for this port.
        'opstatus_reconf': True,
        # If True, configure pipeline if operational status of port changes.
        'receive_lldp': False,
        # If True, receive LLDP on this port.
        'lldp_peer_mac': None,
        # If set, validates src MAC address of incoming LLDP packets
        'max_lldp_lost': 3,
        # threshold before marking a stack port as down
        'dot1x': False,
        # If true, block this port until a successful 802.1x auth
        'dot1x_acl': False,
        # If true, expects authentication and default ACLs for 802.1x auth
        'dot1x_mab': False,
        # If true, allows Mac Auth Bypass on port (NOTE: this is less secure as MACs can be spoofed)
        'dot1x_dyn_acl': False,
        # If true, expects authentication and ACLs with dot1x_assigned flag set
        'restricted_bcast_arpnd': False,
        # If true, this port cannot send non-ARP/IPv6 ND broadcasts to other restricted_bcast_arpnd ports.
        'coprocessor': {},
        # If defined, this port is attached to a packet coprocessor.
        'count_untag_vlan_miss': False,
        # If defined, this port will explicitly count unconfigured native VLAN packets.
    }

    defaults_types = {
        'number': int,
        'name': str,
        'description': str,
        'enabled': bool,
        'permanent_learn': bool,
        'unicast_flood': bool,
        'mirror': (list, str, int),
        'native_vlan': (str, int),
        'tagged_vlans': list,
        'acl_in': (str, int),
        'acls_in': list,
        'stack': dict,
        'max_hosts': int,
        'hairpin': bool,
        'hairpin_unicast': bool,
        'lacp': int,
        'lacp_active': bool,
        'lacp_collect_and_distribute': bool,
        'lacp_unselected': bool,
        'lacp_selected': bool,
        'lacp_standby': bool,
        'lacp_passthrough': list,
        'lacp_resp_interval': int,
        'lacp_port_priority': int,
        'lacp_port_id': int,
        'loop_protect': bool,
        'loop_protect_external': bool,
        'output_only': bool,
        'lldp_beacon': dict,
        'opstatus_reconf': bool,
        'receive_lldp': bool,
        'lldp_peer_mac': str,
        'dot1x': bool,
        'dot1x_acl': bool,
        'dot1x_mab': bool,
        'dot1x_dyn_acl': bool,
        'max_lldp_lost': int,
        'restricted_bcast_arpnd': bool,
        'coprocessor': dict,
        'count_untag_vlan_miss': bool,
    }

    stack_defaults_types = {
        'dp': str,
        'port': (str, int),
    }

    lldp_beacon_defaults_types = {
        'enable': bool,
        'org_tlvs': list,
        'system_name': str,
        'port_descr': str,
    }

    lldp_org_tlv_defaults_types = {
        'oui': (int, bytearray),
        'subtype': (int, bytearray),
        'info': (str, bytearray)
    }

    coprocessor_defaults_types = {
        'strategy': str,
        'vlan_vid_base': int,
    }

    def __init__(self, _id, dp_id, conf=None):
        self.acl_in = None
        self.acls_in = None
        self.description = None
        self.dot1x = None
        self.dot1x_acl = None
        self.dot1x_mab = None
        self.dot1x_dyn_acl = None
        self.dp_id = None
        self.enabled = None
        self.hairpin = None
        self.hairpin_unicast = None
        self.lacp = None
        self.lacp_active = None
        self.lacp_collect_and_distribute = None
        self.lacp_unselected = None
        self.lacp_selected = None
        self.lacp_standby = None
        self.lacp_passthrough = None
        self.lacp_resp_interval = None
        self.lacp_port_priority = None
        self.lacp_port_id = None
        self.loop_protect = None
        self.loop_protect_external = None
        self.max_hosts = None
        self.max_lldp_lost = None
        self.mirror = None
        self.name = None
        self.native_vlan = None
        self.number = None
        self.opstatus_reconf = None
        self.output_only = None
        self.permanent_learn = None
        self.receive_lldp = None
        self.lldp_peer_mac = None
        self.stack = {}
        self.unicast_flood = None
        self.restricted_bcast_arpnd = None
        self.coprocessor = {}
        self.count_untag_vlan_miss = None

        self.dyn_dot1x_native_vlan = None
        self.dyn_lacp_up = None
        self.dyn_lacp_updated_time = None
        self.dyn_lacp_last_resp_time = None
        self.dyn_last_ban_time = None
        self.dyn_last_lacp_pkt = None
        self.dyn_last_lldp_beacon_time = None
        self.dyn_lldp_beacon_recv_state = None
        self.dyn_lldp_beacon_recv_time = None
        self.dyn_learn_ban_count = 0
        self.dyn_phys_up = False
        self.dyn_update_time = None
        self.dyn_stack_current_state = STACK_STATE_NONE
        self.dyn_lacp_port_selected = LACP_PORT_NOTCONFIGURED
        self.dyn_lacp_actor_state = LACP_ACTOR_NOTCONFIGURED
        self.dyn_stack_probe_info = {}

        self.tagged_vlans = []
        self.lldp_beacon = {}
        super().__init__(_id, dp_id, conf)

        # If the port is mirrored convert single attributes to a array
        if self.mirror and not isinstance(self.mirror, list):
            self.mirror = [self.mirror]

    def __str__(self):
        return f'Port {self.number}'

    def __repr__(self):
        return self.__str__()

    def clone_dyn_state(self, prev_port):
        if prev_port:
            self.dyn_lldp_beacon_recv_time = prev_port.dyn_lldp_beacon_recv_time
            self.dyn_lldp_beacon_recv_state = prev_port.dyn_lldp_beacon_recv_state
            self.dyn_stack_current_state = prev_port.dyn_stack_current_state
            self.dyn_last_lldp_beacon_time = prev_port.dyn_last_lldp_beacon_time
            self.dyn_phys_up = prev_port.dyn_phys_up
            self.dyn_stack_probe_info = prev_port.dyn_stack_probe_info
            self.dyn_update_time = prev_port.dyn_update_time

    def stack_descr(self):
        """"Return stacking annotation if this is a stacking port."""
        if self.stack:
            return f'remote DP {self.stack["dp"].name} {self.stack["port"]}'
        return ''

    def set_defaults(self):
        super().set_defaults()
        self._set_default('number', self._id)
        self._set_default('name', str(self._id))
        self._set_default('description', self.name)
        self._set_default('tagged_vlans', [])

    def check_config(self):
        super().check_config()
        test_config_condition(not (isinstance(self.number, int) and self.number > 0 and (
            not valve_of.ignore_port(self.number))), (f'Port number invalid: {self.number}'))
        non_vlan_options = {'stack', 'mirror', 'coprocessor', 'output_only'}
        vlan_agnostic_options = {'enabled', 'number', 'name', 'description', 'max_lldp_lost'}
        vlan_port = self.tagged_vlans or self.native_vlan
        non_vlan_port_options = {option for option in non_vlan_options if getattr(self, option)}
        test_config_condition(
            vlan_port and non_vlan_port_options,
            f'cannot have VLANs configured on non-VLAN ports: {self}')
        if self.output_only:
            test_config_condition(
                not non_vlan_port_options.issubset({'mirror', 'output_only'}),
                f'output_only can only coexist with mirror option on same port {self}')
        elif self.mirror:
            test_config_condition(
                not non_vlan_port_options.issubset({'mirror', 'coprocessor'}),
                f'coprocessor can only coexist with mirror option on same port {self}')
        else:
            test_config_condition(
                len(non_vlan_port_options) > 1,
                f'cannot have multiple non-VLAN port options {non_vlan_port_options} on same port: {self}')
        if non_vlan_port_options:
            for key, default_val in self.defaults.items():
                if key in vlan_agnostic_options or key in non_vlan_port_options:
                    continue
                if key.startswith('acl') and (self.stack or self.coprocessor):
                    continue
                val = getattr(self, key)
                test_config_condition(
                    val != default_val and val,
                    f'Cannot have VLAN option {key}: {val} on non-VLAN port {self}')
        test_config_condition(
            self.hairpin and self.hairpin_unicast,
            'Cannot have both hairpin and hairpin_unicast enabled')
        dot1x_features = {
            dot1x_feature for dot1x_feature, dot1x_config in self.defaults.items()
            if dot1x_feature.startswith('dot1x_') and dot1x_config}
        test_config_condition(
            dot1x_features and not self.dot1x,
            f'802.1x features {dot1x_features} require port to have dot1x enabled')
        if self.dot1x:
            test_config_condition(self.number > 65535, (
                '802.1x not supported on ports > 65535'))
            if self.dot1x_mab:
                test_config_condition(self.dot1x_dyn_acl, (
                    '802.1x_MAB cannot be used with 802.1x_DYN_ACL'))
            if self.dot1x_dyn_acl:
                test_config_condition(self.dot1x_acl, (
                    '802.1x_DYN_ACL cannot be used with 802.1x_ACL'))
        if self.coprocessor:
            self._check_conf_types(self.coprocessor, self.coprocessor_defaults_types)
            test_config_condition(
                self.coprocessor.get('strategy', None) != 'vlan_vid',
                'coprocessor only supports vlan_vid strategy')
            self.coprocessor['vlan_vid_base'] = self.coprocessor.get('vlan_vid_base', 1000)
        if self.stack:
            self._check_conf_types(self.stack, self.stack_defaults_types)
            for stack_config in list(self.stack_defaults_types.keys()):
                test_config_condition(stack_config not in self.stack, (
                    f'stack {stack_config} must be defined'))
            # LLDP always enabled for stack ports.
            self.receive_lldp = True
            if not self.lldp_beacon_enabled():
                self.lldp_beacon.update({'enable': True})
        if self.lacp_resp_interval is not None:
            test_config_condition(
                self.lacp_resp_interval > 65535 or self.lacp_resp_interval < 0.3,
                ('interval must be at least 0.3 and less than 65536'))
        if self.lacp_port_priority is not None:
            test_config_condition(
                self.lacp_port_priority > 255 or self.lacp_port_priority < 0,
                ('lacp port priority must be at least 0 and less than 256'))
        if self.lldp_peer_mac:
            test_config_condition(not netaddr.valid_mac(self.lldp_peer_mac), (
                f'invalid MAC address {self.lldp_peer_mac}'))
        if self.lldp_beacon:
            self._check_conf_types(
                self.lldp_beacon, self.lldp_beacon_defaults_types)
            self.lldp_beacon = self._set_unknown_conf(
                self.lldp_beacon, self.lldp_beacon_defaults_types)
            if self.lldp_beacon_enabled():
                if self.lldp_beacon['port_descr'] is None:
                    self.lldp_beacon['port_descr'] = self.description
                org_tlvs = []
                for org_tlv in self.lldp_beacon['org_tlvs']:
                    self._check_conf_types(org_tlv, self.lldp_org_tlv_defaults_types)
                    test_config_condition(len(org_tlv) != len(self.lldp_org_tlv_defaults_types), (
                        'missing org_tlv config'))
                    if not isinstance(org_tlv['info'], bytearray):
                        try:
                            org_tlv['info'] = bytearray.fromhex(
                                org_tlv['info'])  # pytype: disable=missing-parameter
                        except ValueError:
                            org_tlv['info'] = org_tlv['info'].encode('utf-8')
                    if not isinstance(org_tlv['oui'], bytearray):
                        org_tlv['oui'] = bytearray.fromhex(
                            '%6.6x' % org_tlv['oui'])  # pytype: disable=missing-parameter
                    org_tlvs.append(org_tlv)
                self.lldp_beacon['org_tlvs'] = org_tlvs
        test_config_condition(
            self.acl_in and self.acls_in,
            'Found both acl_in and acls_in, use only acls_in')
        if self.acl_in and not isinstance(self.acl_in, list):
            self.acls_in = [self.acl_in]
            self.acl_in = None
        if self.acls_in:
            for acl in self.acls_in:
                test_config_condition(not isinstance(acl, (int, str)),
                                      'ACL names must be int or str')
        lacp_options = [self.lacp_selected, self.lacp_unselected, self.lacp_standby]
        test_config_condition(
            lacp_options.count(True) > 1,
            'Cannot force multiple LACP port selection states')

    def finalize(self):
        if self.native_vlan:
            test_config_condition(self.native_vlan in self.tagged_vlans, (
                'cannot have same native and tagged VLAN on same port'))
        self.tagged_vlans = tuple(self.tagged_vlans)
        super().finalize()

    def running(self):
        """Return True if port enabled and up."""
        return self.enabled and self.dyn_phys_up

    def vlans(self):
        """Return all VLANs this port is in."""
        if self.native_vlan is not None and self.dyn_dot1x_native_vlan is not None:
            return (self.native_vlan,) + (self.dyn_dot1x_native_vlan,) + tuple(self.tagged_vlans)
        if self.dyn_dot1x_native_vlan is not None:
            return (self.dyn_dot1x_native_vlan,) + tuple(self.tagged_vlans)
        if self.native_vlan is not None:
            return (self.native_vlan,) + tuple(self.tagged_vlans)
        return tuple(self.tagged_vlans)

    def hosts(self, vlans=None):
        """Return all host cache entries this port has learned (on all or specified VLANs)."""
        if vlans is None:
            vlans = self.vlans()
        hosts = []
        for vlan in vlans:
            hosts.extend(list(list(vlan.cached_hosts_on_port(self))))
        return hosts

    def hosts_count(self, vlans=None):
        """Return count of all hosts this port has learned (on all or specified VLANs)."""
        if vlans is None:
            vlans = self.vlans()
        hosts_count = 0
        for vlan in vlans:
            hosts_count += vlan.cached_hosts_count_on_port(self)
        return hosts_count

    def lldp_beacon_enabled(self):
        """Return True if LLDP beacon enabled on this port."""
        return self.lldp_beacon and self.lldp_beacon.get('enable', False)

    def mirror_actions(self):
        """Return OF actions to mirror this port."""
        if self.mirror is not None:
            return [valve_of.output_port(mirror_port) for mirror_port in self.mirror]
        return []

    def non_stack_forwarding(self):
        """Returns True if port is not-stacking and, and able to forward packets."""
        if self.stack:
            return False
        if not self.dyn_phys_up:
            return False
        if self.lacp and not self.dyn_lacp_up:
            return False
        return True

    def tunnel_acls(self):
        """Return any tunnel ACLs on this port."""
        acls = []
        if self.acls_in:
            acls = [acl for acl in self.acls_in if acl.is_tunnel_acl()]
        return acls

    def contains_tunnel_acl(self, tunnel_id=None):
        """Searches through acls_in for a tunnel ACL with a matching tunnel_id"""
        acls = self.tunnel_acls()
        if tunnel_id is None:
            return bool(acls)
        for acl in self.tunnel_acls():
            if acl.get_tunnel_rules(tunnel_id):
                return True
        return False

    # LACP functions
    def lacp_actor_update(self, lacp_up, now=None, lacp_pkt=None, cold_start=False):
        """
        Update the LACP actor state
        Args:
            lacp_up (bool): The intended LACP/port state
            now (float): Current time
            lacp_pkt (PacketMeta): Received LACP packet
            cold_start (bool): Whether the port is being cold started
        Returns:
            current LACP actor state
        """
        self.dyn_lacp_up = 1 if lacp_up else 0
        self.dyn_lacp_updated_time = now
        self.dyn_last_lacp_pkt = lacp_pkt
        if cold_start:
            # Cold starting, so revert to unconfigured LACP state
            self.actor_notconfigured()
        elif not self.running():
            # Phys not up so we do not initialize actor states
            self.actor_none()
        elif not lacp_pkt:
            # Intialize states
            self.actor_init()
        else:
            # Packets received from actor
            if lacp_up:
                # Receiving packets & LACP is UP
                self.actor_up()
            else:
                # Receiving packets but LACP sync bit is not set
                self.actor_nosync()
        return self.actor_state()

    def lacp_port_update(self, selected, cold_start=False):
        """
        Updates the LACP port selection state
        Args:
            selected (bool): Whether the port's DPID is the selected one
            cold_start (bool): Whether the port is being cold started
        Returns
            current lacp port state
        """
        if cold_start:
            # Cold starting, so revert to unconfigured state
            self.deconfigure_port()
        elif self.lacp_selected:
            # Configured LACP port option to be forced into SELECTED state
            self.select_port()
        elif self.lacp_unselected:
            # Configured LACP port option to be forced into UNSELECTED state
            self.deselect_port()
        else:
            if selected:
                # Port SELECTED so change state to SELECTED
                self.select_port()
            elif self.lacp_standby:
                # Send port to STANDBY if not SELECTED
                self.standby_port()
            else:
                # Doesn't belong on chosen DP for LAG, DESELECT port
                self.deselect_port()
        return self.lacp_port_state()

    def get_lacp_flags(self):
        """
        Get the LACP flags for the state the port is in
        Return sync, collecting, distributing flag values
        """
        if self.lacp_collect_and_distribute:
            return 1, 1, 1
        if self.is_port_standby():
            return 1, 0, 0
        if self.is_port_selected():
            return 1, 1, 1
        return 0, 0, 0

    # LACP ACTOR STATES:
    def is_actor_up(self):
        """Return true if the LACP actor state is UP"""
        return self.dyn_lacp_actor_state == LACP_ACTOR_UP

    def is_actor_nosync(self):
        """Return true if the LACP actor state is NOSYNC"""
        return self.dyn_lacp_actor_state == LACP_ACTOR_NOSYNC

    def is_actor_init(self):
        """Return true if the LACP actor state is INIT"""
        return self.dyn_lacp_actor_state == LACP_ACTOR_INIT

    def is_actor_none(self):
        """Return true if the LACP actor state is NONE"""
        return self.dyn_lacp_actor_state == LACP_ACTOR_NONE

    def actor_state(self):
        """Return the current LACP actor state"""
        return self.dyn_lacp_actor_state

    def actor_notconfigured(self):
        """Set the LACP actor state to NOTCONFIGURED"""
        self.dyn_lacp_actor_state = LACP_ACTOR_NOTCONFIGURED

    def actor_none(self):
        """Set the LACP actor state to NONE"""
        self.dyn_lacp_actor_state = LACP_ACTOR_NONE

    def actor_init(self):
        """Set the LACP actor state to INIT"""
        self.dyn_lacp_actor_state = LACP_ACTOR_INIT

    def actor_up(self):
        """Set the LACP actor state to UP"""
        self.dyn_lacp_actor_state = LACP_ACTOR_UP

    def actor_nosync(self):
        """Set the LACP actor state to NOSYNC"""
        self.dyn_lacp_actor_state = LACP_ACTOR_NOSYNC

    @staticmethod
    def actor_state_name(state):
        """Return the string of the actor state"""
        return LACP_ACTOR_DISPLAY_DICT[state]

    # LACP PORT ROLES:
    def is_port_selected(self):
        """Return true if the lacp is a SELECTED port"""
        return self.dyn_lacp_port_selected == LACP_PORT_SELECTED

    def is_port_unselected(self):
        """Return true if the lacp is an UNSELECTED port"""
        return self.dyn_lacp_port_selected == LACP_PORT_UNSELECTED

    def is_port_standby(self):
        """Return true if the lacp is a port in STANDBY"""
        return self.dyn_lacp_port_selected == LACP_PORT_STANDBY

    def lacp_port_state(self):
        """Return the current LACP port state"""
        return self.dyn_lacp_port_selected

    def select_port(self):
        """SELECT the current LACP port"""
        self.dyn_lacp_port_selected = LACP_PORT_SELECTED

    def deselect_port(self):
        """UNSELECT the current LACP port"""
        self.dyn_lacp_port_selected = LACP_PORT_UNSELECTED

    def standby_port(self):
        """Set LACP port state to STANDBY"""
        self.dyn_lacp_port_selected = LACP_PORT_STANDBY

    def deconfigure_port(self):
        """Set LACP port state to NOTCONFIGURED"""
        self.dyn_lacp_port_selected = LACP_PORT_NOTCONFIGURED

    @staticmethod
    def port_role_name(state):
        """Return the LACP port role state name"""
        return LACP_PORT_DISPLAY_DICT[state]

    # STACK PORT ROLES:
    def stack_port_update(self, now):
        """
        Progesses through the stack link state machine

        Args:
            now (float): Current time
        Returns:
            int: Current (new) stack port state
            string: reason for the state change and additional information
        """
        reason = ''
        if self.is_stack_admin_down():
            # Stack port ADMIN_DOWN, so no next state
            return self.stack_state()

        last_seen_lldp_time = self.dyn_stack_probe_info.get('last_seen_lldp_time', None)
        if last_seen_lldp_time is None:
            if self.is_stack_none():
                # New stack, changing to state INIT
                self.stack_init()
                reason = 'new'
        else:
            # Not a new stack port, so progess through state machine
            peer_dp = self.stack['dp']  # pytype: disable=attribute-error
            stack_correct = self.dyn_stack_probe_info.get(
                'stack_correct', None)
            send_interval = peer_dp.lldp_beacon.get(
                'send_interval', peer_dp.DEFAULT_LLDP_SEND_INTERVAL)
            time_since_lldp_seen = None
            num_lost_lldp = None
            stack_timed_out = True

            time_since_lldp_seen = now - last_seen_lldp_time
            num_lost_lldp = time_since_lldp_seen / send_interval
            if num_lost_lldp < self.max_lldp_lost:
                stack_timed_out = False

            if stack_timed_out:
                # Stack timed out, too many packets lost
                self.stack_gone()
                reason = f'too many ({num_lost_lldp}) packets lost, last received {time_since_lldp_seen}s ago'
            elif not stack_correct:
                # Stack bad due to incorrect cabling
                self.stack_bad()
                reason = 'incorrect cabling'
            elif not self.is_stack_up():
                # Nothing gone wrong, so Stack UP
                self.stack_up()
                reason = 'up'
        return self.stack_state(), reason

    def is_stack_admin_down(self):
        """Return True if port is in ADMIN_DOWN state."""
        return self.dyn_stack_current_state == STACK_STATE_ADMIN_DOWN

    def is_stack_none(self):
        """Return True if port is in NONE state."""
        return self.dyn_stack_current_state == STACK_STATE_NONE

    def is_stack_init(self):
        """Return True if port is in INIT state."""
        return self.dyn_stack_current_state == STACK_STATE_INIT

    def is_stack_bad(self):
        """Return True if port is in BAD state."""
        return self.dyn_stack_current_state == STACK_STATE_BAD

    def is_stack_up(self):
        """Return True if port is in UP state."""
        return self.dyn_stack_current_state == STACK_STATE_UP

    def is_stack_gone(self):
        """Return True if port is in GONE state."""
        return self.dyn_stack_current_state == STACK_STATE_GONE

    def stack_state(self):
        """Return the current port stack state"""
        return self.dyn_stack_current_state

    def stack_admin_down(self):
        """Change the current stack state to ADMIN_DOWN."""
        self.dyn_stack_current_state = STACK_STATE_ADMIN_DOWN

    def stack_init(self):
        """Change the current stack state to INIT_DOWN."""
        self.dyn_stack_current_state = STACK_STATE_INIT

    def stack_bad(self):
        """Change the current stack state to BAD."""
        self.dyn_stack_current_state = STACK_STATE_BAD

    def stack_up(self):
        """Change the current stack state to UP."""
        self.dyn_stack_current_state = STACK_STATE_UP

    def stack_gone(self):
        """Change the current stack state to GONE."""
        self.dyn_stack_current_state = STACK_STATE_GONE

    @staticmethod
    def stack_state_name(state):
        """Return stack state name"""
        return STACK_DISPLAY_DICT[state]



#!/usr/bin/env python3

"""Unit tests run as PYTHONPATH=../../.. python3 ./test_valveapp_smoke.py."""

# pylint: disable=protected-access

# Copyright (C) 2015 Research and Innovation Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple
import os
import unittest
from prometheus_client import CollectorRegistry
from os_ken.controller import dpset
from os_ken.controller.ofp_event import EventOFPMsgBase
from faucet import faucet


class OSKenAppSmokeTest(unittest.TestCase):  # pytype: disable=module-attr
    """Test bare instantiation of controller classes."""

    @staticmethod
    def _fake_dp():
        datapath = namedtuple('datapath', ['id', 'close'])(0, lambda: None)
        return datapath

    def test_faucet(self):
        """Test FAUCET can be initialized."""
        os.environ['FAUCET_CONFIG'] = '/dev/null'
        os.environ['FAUCET_LOG'] = '/dev/null'
        os.environ['FAUCET_EXCEPTION_LOG'] = '/dev/null'
        os_ken_app = faucet.Faucet(
            dpset={},
            reg=CollectorRegistry())
        os_ken_app.reload_config(None)
        self.assertFalse(os_ken_app._config_files_changed())
        os_ken_app.metric_update(None)
        event_dp = dpset.EventDPReconnected(dp=self._fake_dp())
        for enter in (True, False):
            event_dp.enter = enter
            os_ken_app.connect_or_disconnect_handler(event_dp)
        for event_handler in (
                os_ken_app.error_handler,
                os_ken_app.features_handler,
                os_ken_app.packet_in_handler,
                os_ken_app.desc_stats_reply_handler,
                os_ken_app.port_desc_stats_reply_handler,
                os_ken_app.port_status_handler,
                os_ken_app.flowremoved_handler,
                os_ken_app.reconnect_handler,
                os_ken_app._datapath_connect,
                os_ken_app._datapath_disconnect):
            msg = namedtuple('msg', ['datapath'])(self._fake_dp())
            event = EventOFPMsgBase(msg=msg)
            event.dp = msg.datapath
            event_handler(event)
        os_ken_app._check_thread_exception()
        os_ken_app._thread_jitter(1)


if __name__ == "__main__":
    unittest.main()  # pytype: disable=module-attr
